
/*
* Swap_Chain.cpp includes the implementations of the functions decleared in Swapping_Chain.h.
*/

//#include <float.h>
#include "Swap_Chain.h"



/*
 * cmp_dist acts as the comparison function of quick_sort.
 */
inline int cmp_dist(	const void* fir, const void* sec)
{
	return (*((p_dist_t**)fir))->dist > (*((p_dist_t**)sec))->dist ? 1 : 
	(*((p_dist_t**)fir))->dist < (*((p_dist_t**)sec))->dist ? -1 : 0;
}

/*
* calc_dist calculates the distance between two objects.
*
* @obj1, obj2 indicate the two objects, respectively.
* @dim indicates the dimension of the object.
*/
DOUBLE calc_dist( obj_t* obj1, obj_t* obj2, int dim)
{
	int i;
	DOUBLE d = 0;
	for( i=0; i<dim; i++)
	{
		d += (DOUBLE)pow( obj2->axis[i]-obj1->axis[i], 2);
	}

	return ( DOUBLE)sqrt( d);
}

/*
 *
 */
context_t* context_alloc( )
{
	context_t* context_v;

	context_v = ( context_t*)malloc( sizeof( context_t));
	memset( context_v, 0, sizeof( context_t));

	//s
	swap_stat_v.memory_v += sizeof( context_t);
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	return context_v;
}

/*
 * context_alloc_part allocates the main memory for the list arrays.
 *
 * @data_v indicates the raw data.
 *
 * @Return the context_t data including the list info.
 */
void context_alloc_part( context_t* context_v)
{
	int i;
	
	data_t* data_v;

	data_v = context_v->data_v;

	//Allocate the main memory for the context_t data.
	//context_v = ( context*)malloc( sizeof( context_t));
	//memset( context_v, 0, sizeof( context_t));

	context_v->m_list = ( m_head_t*)malloc( sizeof( m_head_t) * data_v->n_p_objs);
	memset( context_v->m_list, 0, sizeof( m_head_t) * data_v->n_p_objs);

	//Initialize the variable info.
	context_v->p_objs_num = data_v->n_p_objs;
	context_v->o_objs_num = data_v->n_o_objs;

	//Initialize the heads of the matching list.
	for( i=0; i<data_v->n_p_objs; i++)
	{
		context_v->m_list[ i].head = ( m_node_t*)malloc( sizeof( m_node_t));
		memset( context_v->m_list[ i].head, 0, sizeof( m_node_t));

		context_v->m_list[ i].rear = context_v->m_list[ i].head;

		context_v->m_list[ i].cap = data_v->p_objs[i].weight;
		context_v->m_list[ i].avai = data_v->p_objs[i].weight;
	}

	//s
	swap_stat_v.memory_v += sizeof( m_head_t) * data_v->n_p_objs +
							data_v->n_p_objs * sizeof( m_node_t);
	if( swap_stat_v.memory_v > swap_stat_v.memory_max)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	//
	if( graph_impl_tag == 1)
	{
		context_v->g_list = ( g_head_t*)malloc( sizeof( g_head_t) * data_v->n_o_objs);
		memset( context_v->g_list, 0, sizeof( g_head_t) * data_v->n_o_objs);

		for( i=0; i<data_v->n_o_objs; i++)
		{
			context_v->g_list[ i].head = ( g_node_t*)malloc( sizeof( g_node_t));
			memset( context_v->g_list[ i].head, 0, sizeof( g_node_t));
		}

		//s
		swap_stat_v.memory_v += sizeof( g_head_t) * data_v->n_o_objs +
								data_v->n_o_objs * sizeof( g_node_t);
		if( swap_stat_v.memory_v > swap_stat_v.memory_max)
			swap_stat_v.memory_max = swap_stat_v.memory_v;
		//s
	}
}

/*
* random_ini_match_list initializes the matching list info without any heuristic.
*
* @context_v indicates the context info.
* @data_v indicates the raw data info.
*/
void random_ini_match_list( context_t* context_v)
{
	int i, p_id, cur;
	float n_pair;
	FLOAT  o_wei, p_wei;
	m_node_t* m_node_v;

	data_t* data_v;

	//
	data_v = context_v->data_v;

	n_pair = 0;
	cur = 0;

	o_wei = data_v->o_objs[ cur].weight;
	p_wei = data_v->p_objs[ cur].weight;

	for( i=0; i<data_v->n_o_objs; i++)
	{
		m_node_v = ( m_node_t*)malloc( sizeof( m_node_t));
		memset( m_node_v, 0, sizeof( m_node_t));

		m_node_v->id = i;
		m_node_v->dist = calc_dist( data_v->o_objs + i, data_v->p_objs + cur, data_v->dim);
		
		p_id = cur;
		//rear = context_v->m_list[ p_id].head;

		n_pair ++;

		//s
		swap_stat_v.memory_v += sizeof( m_node_t);
		//s

		if( o_wei > p_wei)
		{
			m_node_v->wei = p_wei;
			o_wei -= p_wei;

			//Updated on 27 Feb by Cheng.
			cur ++;
			i --;
		}
		else	//o_wei <= p_wei;
		{
			m_node_v->wei = o_wei;
			p_wei -= o_wei;

			if( p_wei == 0)
				cur ++;

			/*
			if( i != data_v->n_o_objs-1)
			{
				if( p_wei == 0)
					p_wei = data_v->p_objs[ ++cur].weight;

				o_wei = data_v->o_objs[ i+1].weight;
			}
			*/
		}

		//
		context_v->m_list[ p_id].rear->next = m_node_v;
		context_v->m_list[ p_id].rear = m_node_v;
		
		context_v->m_list[ p_id].avai -= m_node_v->wei;

		if( context_v->m_list[ p_id].max_d == NULL || 
			context_v->m_list[ p_id].max_d->dist < m_node_v->dist)
			context_v->m_list[ p_id].max_d = m_node_v;

	}//for(i)

	//s
	swap_stat_v.n_pair_v = n_pair;
	swap_stat_v.n_ini_pair = n_pair;
	if( swap_stat_v.n_pair_max < n_pair)
		swap_stat_v.n_pair_max = n_pair;
	
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s
}

/*
* sort_ini_match_list initializes the matching list corresponding to "sort_ini_context".
*
* @Interfaces same to those of "ini_match_list".
*/
void sort_ini_match_list( context_t* context_v)
{
	int i, j, p_id, b_tag;
	float n_pair;
	FLOAT o_dem;
	m_node_t* m_node_v;

	p_dist_t*	p_dist_v;
	p_dist_t**	o_dist_v;

	data_t* data_v;

	//
	data_v = context_v->data_v;

	p_dist_v = ( p_dist_t*)malloc( sizeof( p_dist_t) * data_v->n_p_objs);
	memset( p_dist_v, 0, sizeof( p_dist_t) * data_v->n_p_objs);

	o_dist_v = ( p_dist_t**)malloc( sizeof( p_dist_t*) * data_v->n_p_objs);
	memset( o_dist_v, 0, sizeof( p_dist_t*) * data_v->n_p_objs);

	//s
	swap_stat_v.memory_v += sizeof( p_dist_t) * data_v->n_p_objs +
							sizeof( p_dist_t*) * data_v->n_p_objs;
	//s

	for( i=0; i<data_v->n_p_objs; i++)
	{
		p_dist_v[i].p_id = i;
		p_dist_v[i].avai = data_v->p_objs[i].weight;
	}

	//
	n_pair = 0;
	for( i=0; i<data_v->n_o_objs; i++)
	{
		//Construct the distance array for "o[i]".
		for( j=0; j<data_v->n_p_objs; j++)
		{
			p_dist_v[j].dist = calc_dist( data_v->o_objs+i, data_v->p_objs+j, data_v->dim);
			o_dist_v[j] = &p_dist_v[j];
		}

		//Sorting the distance array.
		qsort( o_dist_v, data_v->n_p_objs, sizeof(p_dist_t*), cmp_dist);

		//Making the initial matchings.
		b_tag = 0;
		o_dem=data_v->o_objs[i].weight;

		for( j=0; j<data_v->n_p_objs; j++)
		{
			if( o_dist_v[j]->avai == 0)
				continue;

			m_node_v = ( m_node_t*)malloc( sizeof( m_node_t));
			memset( m_node_v, 0, sizeof( m_node_t));

			m_node_v->dist = o_dist_v[ j]->dist;
			m_node_v->id = i;

			//s
			swap_stat_v.memory_v += sizeof( m_node_t);
			//s

			n_pair ++;

			if( o_dem < o_dist_v[j]->avai)
			{
				m_node_v->wei = o_dem;
				o_dist_v[j]->avai -= m_node_v->wei;

				//o_dem = 0;

				b_tag = 1;
			}
			else	//o_dem >= o_dist_v[j]->avai
			{
				if( o_dem == o_dist_v[j]->avai)
					b_tag = 1;

				m_node_v->wei = o_dist_v[j]->avai;
				o_dist_v[j]->avai = 0;

				o_dem -= m_node_v->wei;
			}

			//Consider the order when adding matching pair into the m_list.
			p_id = o_dist_v[j]->p_id;
			context_v->m_list[p_id].avai -= m_node_v->wei;

			
			context_v->m_list[ p_id].rear->next = m_node_v;
			context_v->m_list[ p_id].rear = m_node_v;

			if( context_v->m_list[p_id].max_d == NULL || 
				context_v->m_list[p_id].max_d->dist < m_node_v->dist)
				context_v->m_list[p_id].max_d = m_node_v;

			if( b_tag == 1)
				break;
		}//for(j)

//t
		if( j == data_v->n_p_objs && b_tag != 1)
		{
			fprintf( stderr, "[sort_ini_match_list] The service providers cannot satisfy all the customers.\n");
			exit( 0);
		}
//t

	}//for(i)


	free( p_dist_v);
	free( o_dist_v);

	//s
	swap_stat_v.n_pair_v = n_pair;
	swap_stat_v.n_ini_pair = n_pair;
	if( swap_stat_v.n_pair_max < n_pair)
		swap_stat_v.n_pair_max = n_pair;

	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;

	//Release the memory.
	swap_stat_v.memory_v -= sizeof( p_dist_t) * data_v->n_p_objs +
							sizeof( p_dist_t*) * data_v->n_p_objs;
	
	//s
}

/*
* Construct_R_Tree constructs a R-tree based on the P_points.
*
void Construct_R_Tree( )
{
	int i, j;
	object* obj;

	//Initialize the related variables related to the tree.
	Initial( 2);

	//Construct the R_tree by inserting the points one by one.
	for( i=0; i<data_v->n_p_objs; i++)
	{
		obj = (object*)malloc( sizeof(object));
		obj->rect = (range_t*)malloc( sizeof( range_t) * data_v->dim);
		obj->id = i;
		obj->avai = data_v->p_objs[i].weight;
		for( j=0; j<data_v->dim; j++)
			obj->rect[j].max = obj->rect[j].min = data_v->p_objs[i].axis[j];

		RStaInsert( obj, 0);
		rTree.number ++;
	}	
}

/*
* kNN_ini_match_list the matching list corresponding to "sort_ini_context".
*
* @Interfaces same to those of "ini_match_list".
*
void kNN_ini_match_list( context_t* context_v)
{
	int i, j, p_id, b_tag;
	float n_pair;
	long r_tree_size = 0, kNN_size = 0;
	FLOAT o_dem;
	m_node_t* m_node_v1, *m_node_v2;
	point p_point_v;
	kNearestObj* kNN_v;

	//Initialize the kNN related variables.
	p_point_v.ord = (float*)malloc( sizeof(float) * data_v->dim);
	kNN_v = (kNearestObj*)malloc( sizeof(kNearestObj));
	kNN_v->cur = 0;
	kNN_v->size = 1;
	kNN_v->obj = (object**)malloc( sizeof(object*) * 1);

	kNN_size += sizeof(float) * data_v->dim + sizeof(kNearestObj) + sizeof(object*);

	//Construct a r-tree for P.
	Construct_R_Tree( );

#ifndef WIN32	
	get_cur_time( &t_rTree_end);
#endif

	/*s*
	r_tree_size += rTree.number * sizeof( node) + 
		rTree.number * sizeof(range_t) * data_v->dim * M +
		rTree.number * sizeof(void*) * M +
		data_v->n_p_objs * sizeof(object) +
		data_v->n_p_objs * sizeof(range_t) * data_v->dim;

	swap_stat_v.memory_v += kNN_size + r_tree_size;
	/*s*

	//Making initial matchings.
	for( i=0; i<data_v->n_o_objs; i++)
	{
		for( j=0; j<data_v->dim; j++)
			p_point_v.ord[j] = data_v->o_objs[i].axis[j];
		o_dem = data_v->o_objs[i].weight;

		while( true)
		{
			kNN_v->nearest = DBL_MAX;
			kNN_v->cur = 0;

			NN( rTree.root, &p_point_v, kNN_v);
/*t*
			if( kNN_v->obj[0]->avai == 0)
			{
				fprintf( stderr, "The node should have been deleted.\n");
				exit(0);
			}
/*t*

			m_node_t* m_node_v = (m_node_t*)malloc( sizeof( m_node_t));
			m_node_v->id = i;
			m_node_v->dist = kNN_v->nearest;	

			/*s*
			swap_stat_v.memory_v += sizeof( m_node_t);
			/*s*

			n_pair ++;

			if( o_dem < kNN_v->obj[0]->avai)
			{
				m_node_v->wei = o_dem;
				kNN_v->obj[0]->avai -= m_node_v->wei;
				b_tag = 1;
			}
			else
			{
				if( o_dem == kNN_v->obj[0]->avai)
					b_tag = 1;
				m_node_v->wei = kNN_v->obj[0]->avai;
				o_dem -= m_node_v->wei;
				kNN_v->obj[0]->avai = 0;
			}

			//Consider the order when adding matching pair into the m_list.
			p_id = kNN_v->obj[0]->id;
			context_v->m_list[p_id].avai -= m_node_v->wei;

			if( context_v->m_list[p_id].next == NULL ||
				m_node_v->id < context_v->m_list[p_id].next->id)		//boundary case.
			{
				m_node_v->next = context_v->m_list[p_id].next;
				context_v->m_list[p_id].next = m_node_v;
			}
			else												//normal case.
			{
				m_node_v1 = context_v->m_list[p_id].next;
				while( m_node_v1 != NULL && m_node_v1->id < m_node_v->id)
				{
					m_node_v2 = m_node_v1;
					m_node_v1 = m_node_v1->next;
				}

				m_node_v2->next = m_node_v;
				m_node_v->next = m_node_v1;
			}

			if( context_v->m_list[p_id].max_d == NULL || 
				context_v->m_list[p_id].max_d->dist < m_node_v->dist)
				context_v->m_list[p_id].max_d = m_node_v;

			if( kNN_v->obj[0]->avai == 0)
			{
				if( !Delete( kNN_v->obj[0]))
				{
					fprintf( stderr, "The deletion node doesn't exist in the R_tree.\n");
					exit(0);
				}
			}

			if( b_tag == 1)
				break;
		}//while.
	}//for(i)

	//Release the resource.
	free( p_point_v.ord);
	free( kNN_v->obj);
	free( kNN_v);
	//Free the r_tree--to be finished.

	/*s*
	swap_stat_v.memory_v -= kNN_size + r_tree_size;
	swap_stat_v.n_pair_v = n_pair;
	swap_stat_v.n_ini_pair = n_pair;
	if( swap_stat_v.n_pair_max < n_pair)
		swap_stat_v.n_pair_max = n_pair;

	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	/*s*
}
*/


/*
 *	This option is removed due to its inefficiency.
 */
void kNN_ini_match_list( context_t* context_v)
{
	//
}

/*
* fair_ini_match_list initializes the the matching list using fair algorithm.
*/
void fair_ini_match_list( context_t* context_v)
{
	int i, n_match, p_id;
	m_node_t* m_node_v, *m_node_v1;

	data_t* data_v;

	//
	data_v = context_v->data_v;

	assignment* assi_v = comp_fair_matching( context_v->data_v);
	n_match = assi_v->noOfMatch;

	for( i=0; i<n_match; i++)
	{
		//Creating a matching pair.
		m_node_v = (m_node_t*)malloc( sizeof( m_node_t));
		memset( m_node_v, 0, sizeof( m_node_t));

		m_node_v->id = assi_v->list[ i]->oid;
		m_node_v->wei = ( float)assi_v->list[ i]->weight;
		m_node_v->dist = calc_dist( data_v->p_objs+assi_v->list[i]->pid, data_v->o_objs+assi_v->list[i]->oid, data_v->dim);

		//s
		swap_stat_v.memory_v += sizeof( m_node_t);
		//s
		
		//Inserting the matching pair into the matching list(order).
		p_id = assi_v->list[ i]->pid;
		context_v->m_list[ p_id].avai -= m_node_v->wei;

		m_node_v1 = context_v->m_list[ p_id].head;
		while( m_node_v1->next != NULL && m_node_v1->next->id < m_node_v->id)
			m_node_v1 = m_node_v1->next;

		m_node_v->next = m_node_v1->next;
		m_node_v1->next = m_node_v;

		if( m_node_v1 == context_v->m_list[ p_id].rear)
			context_v->m_list[ p_id].rear = m_node_v;
		
		if( context_v->m_list[p_id].max_d == NULL || 
			context_v->m_list[p_id].max_d->dist < m_node_v->dist)
			context_v->m_list[p_id].max_d = m_node_v;
	}//

	//s
	swap_stat_v.n_pair_v = (float)assi_v->noOfMatch;
	swap_stat_v.n_ini_pair = (float)assi_v->noOfMatch;
	if( swap_stat_v.n_pair_max < (float)assi_v->noOfMatch)
		swap_stat_v.n_pair_max = (float)assi_v->noOfMatch;
	
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	//Releasing the resource related the assignment.
	for( i=0; i<assi_v->noOfMatch; i++)
		free( assi_v->list[i]);
	free( assi_v);
}



/*
* ini_context_member initializes the members of the context structure.
*
* @context_v indicates the context data info.
*/
void ini_context_member( context_t* context_v)
{
	int i, p_id;
	DOUBLE max_d = -1;

	//Searching for the current "mmd" value.
	for( i=0; i<context_v->p_objs_num; i++)
	{
		if( context_v->m_list[i].max_d != NULL && context_v->m_list[i].max_d->dist > max_d)
		{
			p_id = i;
			//o_m_id = context_v->m_list[i].max_d->id;
			max_d = context_v->m_list[i].max_d->dist;
		}
	}

	//Updated on 27 Feb. 2013.
	//t
	if( max_d < 0)
	{
		fprintf( stderr, "[ini_context_member] Cannot find the mmd.\n");
		exit( 0);
	}
	//t

	context_v->mmd = max_d;
	context_v->p_id = p_id;

	context_v->o_m = context_v->m_list[ p_id].max_d->id;
	context_v->o_m_dem = 0;
}

/*
* ini_graph_list initializes the graph list info by remaining the edges smaller than mmd.
*
* @context_v, data_v same to those of "ini_match_list".
* @mmd indicates the mmd value.
*/
void ini_graph_list( context_t* context_v, DOUBLE mmd)
{
	int i, j;
	float n_edge;
	DOUBLE dist_v;
	g_node_t* g_node_v;

	data_t* data_v;

	//
	data_v = context_v->data_v;


	n_edge = 0;

	for( i=0; i<data_v->n_o_objs; i++)
	{
		for( j=0; j<data_v->n_p_objs; j++)
		{
			dist_v = calc_dist( data_v->o_objs+i, data_v->p_objs+j, data_v->dim);

			if( dist_v <= mmd)
			{
				n_edge ++;
					
				g_node_v = (g_node_t*)malloc( sizeof( g_node_t));
				memset( g_node_v, 0, sizeof( g_node_t));

				//s
				swap_stat_v.memory_v += sizeof( g_node_t);
				//s

				g_node_v->dist = dist_v;
				g_node_v->id = j;

				g_node_v->next = context_v->g_list[ i].head->next;
				context_v->g_list[ i].head->next = g_node_v;
			}
		}//for(j)
	}//for(i)

	//s
	swap_stat_v.n_ini_edge = n_edge;
	swap_stat_v.n_prev_edge = n_edge;
	swap_stat_v.n_edge_sum += ( float)n_edge;

	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s
}


/*
* ini_context initializes the context_t variable based on the raw data information *randomly*.
*
* @data_v indicates the raw data information.
*
* @Return the initialized the context data.
*
void random_ini_context( context_t* context_v)
{
	//Initialize the matching list information.
	random_ini_match_list( context_v);

	//Initialize the members of the context data.
	ini_context_member( context_v);
	
	//Initialize the graph list information.
	if( graph_impl_tag == 1)
		ini_graph_list( context_v, context_v->mmd);	
}*/

/*
* sort_ini_context initializes the context data using the heuristic of "nearest neighbor".(Sorting)
*
* @Same interfaces to those of "random_ini_context".
*
void sort_ini_context( context_t* context_v)
{
	//Initialize the matching list.
	sort_ini_match_list(context_v);

	//Initialize the members of the context data.
	ini_context_member( context_v);
	
	//Initialize the graph list information.
	if( graph_impl_tag == 1)
		ini_graph_list( context_v, context_v->mmd);	
}*/

/*
* kNN_ini_match_list initializes the context data using the heuristic of k-NN.(R_tree/R_star_tree)
*
* @Same interfaces to those of the "random_ini_context".
*
void kNN_ini_context( context_t* context_v)
{
	//Initialize the matching list.
	kNN_ini_match_list(context_v);

	//Initialize the members of the context data.
	ini_context_member( context_v);
	
	//Initialize the graph list information.
	if( graph_impl_tag == 1)
		ini_graph_list( context_v, context_v->mmd);
}*/

/*
* fair_ini_context initializes the Matching list using fair method.
*
void fair_ini_context( context_t* context_v)
{
	//Initializing the matching list.
	fair_ini_match_list(context_v);

	//Initializing the members of the context data.
	ini_context_member( context_v);
	
	//Initializing the graph list information.
	if( graph_impl_tag == 1)
		ini_graph_list( context_v, context_v->mmd);	
}*/

/*
* context_release releases the related resources to context data.
*
* @context_v indicates the context data.
*/
void context_release( context_t* context_v)
{
	int i;
	m_node_t* m_node_v1, *m_node_v2;
	g_node_t* g_node_v1, *g_node_v2;

	//Release m_list.
	for( i=0; i<context_v->p_objs_num; i++)
	{
		m_node_v1 = context_v->m_list[i].head;
		while( m_node_v1 != NULL)
		{
			m_node_v2 = m_node_v1->next;
			free( m_node_v1);
			m_node_v1 = m_node_v2;

			//s
			swap_stat_v.memory_v -= sizeof( m_node_t);
			//s
		}
	}

	free( context_v->m_list);

	//s
	swap_stat_v.memory_v -= sizeof( m_head_t) * context_v->p_objs_num;
	//s

	//Release g_list.
	if( graph_impl_tag == 1)
	{
		for( i=0; i<context_v->o_objs_num; i++)
		{
			g_node_v1 = context_v->g_list[i].head;
			while( g_node_v1 != NULL)
			{
				g_node_v2 = g_node_v1->next;
				free( g_node_v1);
				g_node_v1 = g_node_v2;
				
				//s
				swap_stat_v.memory_v -= sizeof( g_node_t);
				//s
			}
		}
		
		free( context_v->g_list);

		//s
		swap_stat_v.memory_v -= sizeof( g_head_t) * context_v->o_objs_num;
		//s
	}

	//Release the data_v.
	data_release( context_v->data_v);

	free( context_v);

	//s
	swap_stat_v.memory_v -= sizeof( context_t);
	//s
}

/*
* construct_sc constructs the "swapping chain" based on the BFS searching queue info.
*
* @que indicates the searching queue.
* @rear indicates the end position of the queue.
*
* @Return the swapping chain.
*/
s_chain* construct_sc( q_node_t** que, int rear)
{
	int cur, cnt = 0;
	FLOAT min_amount = FLT_MAX;

	s_chain_t* c = ( s_chain_t*)malloc( sizeof( s_chain_t));
	memset( c, 0, sizeof( s_chain_t));

	//Calculate the info of the "swapping amount" and the length of the chain.
	cur = rear - 1;
	while( cur != -1)
	{
		if( que[ cur]->amount < min_amount)
			min_amount = que[ cur]->amount;
		
		cnt ++;
		
		cur = que[ cur]->prev;
	};
	
	//Construct the "swapping chain".
	c->amount = min_amount;
	c->leng = cnt;
	c->chain = ( int*)malloc( sizeof( int) * cnt);
	memset( c->chain, 0, sizeof( int) * cnt);

	c->dist = (DOUBLE*)malloc( sizeof( DOUBLE) * cnt);
	memset( c->dist, 0, sizeof( DOUBLE) * cnt);

	//s
	swap_stat_v.memory_v += sizeof( s_chain_t) + sizeof( int) * cnt + sizeof( DOUBLE) * cnt;
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	cur = rear - 1;
	while( cnt >= 1)
	{
		c->chain[ --cnt] = que[ cur]->id;
		c->dist[ cnt] = que[ cur]->dist;

		cur = que[ cur]->prev;
	}

	return c;
}

/*
* search_sc_graph searchs for the corresponding the "swapping chain" w.r.t. o_m bounded by mmd included in context_v.
*
* @context_v indicates the information of the graph, matching, o_m, mmd and so on.
*
* @Return the chain if exists; otherwise, return NULL.
*/
s_chain* search_sc_graph( context_t* context_v)
{
	int cur, rear = 0, suc = 0;
	int* p_tag, *o_tag;
	float expension_v = 0;

	q_node_t* q_node_v, *q_node_tmp;
	g_node_t* g_node_v, *g_node_v1;
	m_node_t* m_node_v;

	q_node_t** que;
	s_chain_t* sc;

	//Allocate the main memory.
	que = ( q_node_t**)malloc( sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num));
	memset( que, 0, sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num));

	p_tag = ( int*)malloc( sizeof( int) * context_v->p_objs_num);
	o_tag = ( int*)malloc( sizeof( int) * context_v->o_objs_num);
	memset( p_tag, 0, sizeof( int) * context_v->p_objs_num);
	memset( o_tag, 0, sizeof( int) * context_v->o_objs_num);
	
	//s
	swap_stat_v.memory_v += sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num) +
							sizeof( int) * context_v->p_objs_num + 
							sizeof( int) * context_v->o_objs_num;
	//s

	//Push o_m into the queue.
	q_node_v = ( q_node_t*)malloc( sizeof( q_node_t));
	memset( q_node_v, 0, sizeof( q_node_t));

	q_node_v->id = context_v->o_m;
	q_node_v->tag = 0;
	q_node_v->amount = context_v->o_m_dem;
	q_node_v->dist = 0;
	q_node_v->prev = -1;
	o_tag[ context_v->o_m] = 1;

	que[ rear++] = q_node_v;
	cur = 0;

	//Searching the "swapping chain" using BFS method.
	while( cur != rear)
	{
		q_node_tmp = que[ cur];
		if( q_node_tmp->tag == 0)		//The current node is a O object.
		{
			//Push all the P objects that are within the distance of mmd from the O object being delt with currently.
			g_node_v1 = context_v->g_list[ q_node_tmp->id].head;
			g_node_v = g_node_v1->next;
			while( g_node_v != NULL)
			{
				//s
				expension_v ++;
				//s

				if( g_node_v->dist > context_v->mmd)
				{
					//Update the graph.
					g_node_v1->next = g_node_v->next;

					free( g_node_v);

					//s
					swap_stat_v.memory_v -= sizeof( g_node_t);
					//s

					g_node_v = g_node_v1->next;

					continue;
				}
				else if( p_tag[ g_node_v->id] == 0 && g_node_v->dist < context_v->mmd)	//un-tagged P object.
				{
					q_node_v = (q_node_t*)malloc( sizeof( q_node_t));
					memset( q_node_v, 0, sizeof( q_node_t));

					q_node_v->id = g_node_v->id;
					q_node_v->dist = g_node_v->dist;
					q_node_v->tag = 1;
					q_node_v->amount = q_node_tmp->amount;	//temporary use. INF is OK here.
					q_node_v->prev = cur;
					p_tag[ g_node_v->id] = 1;
					
					que[ rear++] = q_node_v;

					if( context_v->m_list[ g_node_v->id].avai > 0)	//Forming a swapping chain, stop.
					{						
						suc = 1;
						goto end;
					}
				}

				g_node_v1 = g_node_v;
				g_node_v = g_node_v1->next;

			}//while
		}
		else							//The current node is a P object.
		{
			//Push all the O objects that are currently matched with the P object being dealt with.
			m_node_v = context_v->m_list[ q_node_tmp->id].head->next;
			while( m_node_v != NULL)
			{
				//s
				expension_v ++;
				//s

				if( o_tag[ m_node_v->id] == 0)	//un-tagged O object.
				{
					q_node_v = (q_node_t*)malloc( sizeof( q_node_t));
					memset( q_node_v, 0, sizeof( q_node_t));

					q_node_v->id = m_node_v->id;
					//q_node_v->dist = 0;		
					q_node_v->tag = 0;
					q_node_v->amount = m_node_v->wei;
					q_node_v->prev = cur;

					o_tag[ m_node_v->id] = 1;

					que[ rear++] = q_node_v;
				}
				
				m_node_v = m_node_v->next;
		
			}//while
		}//else

		cur ++;
	}//while(cur != rear)

end:
	//s
	swap_stat_v.memory_v += sizeof( q_node_t) * rear;

	if( rear < swap_stat_v.n_swap_que.min)
		swap_stat_v.n_swap_que.min = (float)rear;
	if( rear > swap_stat_v.n_swap_que.max)
		swap_stat_v.n_swap_que.max = (float)rear;
	swap_stat_v.n_swap_que_sum += (float)rear;

	if( expension_v > swap_stat_v.n_expension.max)
		swap_stat_v.n_expension.max = expension_v;
	if( expension_v < swap_stat_v.n_expension.min)
		swap_stat_v.n_expension.min = expension_v;
	swap_stat_v.n_expension_sum += (float)expension_v;
	//s

	if( suc == 1)	//Case 1: A swapping chain has been found.
		sc = construct_sc( que, rear);
	else			//Case 2: No swapping chain exists.
	{
		sc = NULL;
	}

	//s
	if( swap_stat_v.memory_v > swap_stat_v.memory_max)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	//Release the related the resources.
	cur = 0;
	while( cur < rear)
		free( que[ cur++]);
	free( que);
	free( o_tag);
	free( p_tag);

	//s
	swap_stat_v.memory_v -= sizeof( q_node_t) * rear +
							sizeof( q_node_t*) * ( context_v->p_objs_num + context_v->o_objs_num) +
							sizeof( int) * context_v->p_objs_num + 
							sizeof( int) * context_v->o_objs_num;
	//s
	
	return sc;
}

/*
* Locate_Match_Form locates the feasible pair (m_node_t) within the matching list.
*	(Used for "form_match".)
*
* @context_v indicates the spatial data information.
* @p_id, o_id indicate the P object and O object, respectively.
*
* @Return the matching pair identified by a "m_node".
*
m_node_t* Locate_Match_Form( context_t* context_v, int p_id, int o_id)
{
	m_node_t* m_node_v1, *m_node_v2, *m_node_v3;

	m_node_v1 = context_v->m_list[p_id].next;

	//Deal with the boundary cases.
	if( m_node_v1 == NULL || m_node_v1->id > o_id)	//
	{
		m_node_v2 = (m_node_t*)malloc( sizeof( m_node_t));
		memset( m_node_v2, 0, sizeof( m_node_t));
		m_node_v2->dist = -1;

		m_node_v2->next = m_node_v1;
		context_v->m_list[p_id].next = m_node_v2;

		//s
		swap_stat_v.memory_v += sizeof( m_node_t);
		swap_stat_v.n_pair_v ++;

		if( swap_stat_v.n_pair_max < swap_stat_v.n_pair_v)
			swap_stat_v.n_pair_max = swap_stat_v.n_pair_v;
		//s

		return m_node_v2;
	}

	//Deal with the normal cases.
	while( m_node_v1 != NULL && m_node_v1->id < o_id)
	{
		m_node_v2 = m_node_v1;
		m_node_v1 =m_node_v1->next;
	}

	if( m_node_v1 == NULL || m_node_v1->id > o_id)
	{
		m_node_v3 = (m_node_t*)malloc( sizeof( m_node_t));
		memset( m_node_v3, 0, sizeof( m_node_t));
		m_node_v3->dist = -1;

		m_node_v3->next = m_node_v1;
		m_node_v2->next = m_node_v3;

		//s
		swap_stat_v.memory_v += sizeof( m_node_t);
		swap_stat_v.n_pair_v ++;
		
		if( swap_stat_v.n_pair_max < swap_stat_v.n_pair_v)
			swap_stat_v.n_pair_max = swap_stat_v.n_pair_v;	
		//s

		return m_node_v3;
	}
	else	//m_node_v1->id == o_id;
		return m_node_v1;
}
*/

/*
* form_match forms the matching pair indicated by (p_id, o_id, wei).
*	Two cases for this function: 
*		1. forming a new pair; 
*		2. adding additionl matching weight on the existing pair;
*
* @context_v indicates the the spatial data information.
* @(p_id, o_id, wei) indicates the matching pair to be formed.
*
*/
void form_match( context_t* context_v, int p_id, int o_id, FLOAT wei, DOUBLE dist)
{
	m_node_t* m_node_v1, *m_node_v2;
	
	
	//Locate the target match.
	m_node_v1 = context_v->m_list[ p_id].head;
	while( m_node_v1->next != NULL && m_node_v1->next->id < o_id)
		m_node_v1 = m_node_v1->next;

	//Form the match.
	if( m_node_v1->next == NULL || m_node_v1->next->id > o_id)
	{
		//The target match is a new match.
		m_node_v2 = ( m_node_t*)malloc( sizeof( m_node_t));
		memset( m_node_v2, 0, sizeof( m_node_t));

		m_node_v2->id = o_id;
		m_node_v2->dist = dist;
		m_node_v2->wei = wei;

		m_node_v2->next = m_node_v1->next;
		m_node_v1->next = m_node_v2;

		if( context_v->m_list[ p_id].rear == m_node_v1)
			context_v->m_list[ p_id].rear = m_node_v2;

		if( context_v->m_list[ p_id].max_d == NULL ||
			dist > context_v->m_list[ p_id].max_d->dist)
			context_v->m_list[ p_id].max_d = m_node_v2;


		//s
		swap_stat_v.n_pair_v ++;
		if( swap_stat_v.n_pair_max < swap_stat_v.n_pair_v)
			swap_stat_v.n_pair_max = swap_stat_v.n_pair_v;

		swap_stat_v.memory_v += sizeof( m_node_t);
		if( swap_stat_v.memory_max < swap_stat_v.memory_v)
			swap_stat_v.memory_max = swap_stat_v.memory_v;
		//s
	}
	else
	{
		//The target match is an existing match.
		m_node_v1->next->wei += wei;
	}

	//Update the corresponidng list.
	context_v->m_list[ p_id].avai -= wei;
}

/*
* Locate_Match_Break locates the matching pair (m_node_t) within the matching list.
*	(Used for "break_match".)
*
* @context_v indicates the spatial data information.
* @p_id, o_id indicate the P object and O object, respectively.
*
* @Return the neighbor node just before the matching pair identified by a "m_node".
*
m_node_t* Locate_Match_Break( context_t* context_v, int p_id, int o_id)
{
	m_node_t* m_node_v1, *m_node_v2;


	m_node_v1 = context_v->m_list[ p_id].head->next;

//t
	if( m_node_v1 == NULL || m_node_v1->id > o_id)
	{
		fprintf( stderr, "The matching pair (%i,%i) to be broken doesn't exsit.\n", p_id, o_id);
		exit( 0);
	}
//t

	//Deal with the normal cases.
	m_node_v2 = NULL;
	while( m_node_v1 != NULL && m_node_v1->id < o_id)
	{
		m_node_v2 = m_node_v1;
		m_node_v1 =m_node_v1->next;
	}

	if( m_node_v1->id == o_id)
	{
		//Note that 1. m_node_v2 is the node just before m_node_v1--the node to be broken;
		//			2. m_node_v2 can be NULL.
		return m_node_v2;
	}
//t
	else	//m_node_v1->id > o_id OR m_node_v1 == NULL
	{
		fprintf( stderr, "The matching pair (%i,%i) to be broken doesn't exsit.\n", p_id, o_id);
		exit( 0);
	}
//t
}
*/

/*
* update_max_d updates the information of "max_d" of the matching list.
*
* @head indicates the matching list.
*/
void update_max_d( context_t* context_v, int p_id)
{
	DOUBLE max_d;
	m_node_t* m_node_v;

	max_d = -1;
	context_v->m_list[ p_id].max_d = NULL;

	m_node_v = context_v->m_list[ p_id].head->next;
	while( m_node_v != NULL)
	{
		if( m_node_v->dist > max_d)
		{
			max_d = m_node_v->dist;
			context_v->m_list[ p_id].max_d = m_node_v;
		}

		m_node_v = m_node_v->next;
	}
}

/*
* break_match breaks down the matching pair indicated by (p_id, o_id, wei).
*
* @Tnterfaces same to those of "form_match" except the "dist".
*
void break_match( context_t* context_v, int p_id, int o_id, FLOAT wei)
{
	m_node_t* m_node_v, *m_node_v1;

	context_v->m_list[ p_id].avai += wei;
	m_node_v = Locate_Match_Break( context_v, p_id, o_id);
	
	if( m_node_v == NULL)	//Case 1: the matching pair to be broken down is the first m_node in the list.
	{
		//t
		if( context_v->m_list[p_id].next->wei < wei)		//This case NEVER happens.
		{
			fprintf( stderr, "Attempt to break down the matching pair with more than its weight.\n");
			exit( 0);
		}
		//t

		if( context_v->m_list[ p_id].next->wei == wei)		//The m_node should be deleted.
		{
			m_node_v = context_v->m_list[p_id].next;
			context_v->m_list[p_id].next = context_v->m_list[p_id].next->next;
			free( m_node_v);

			//s
			swap_stat_v.memory_v -= sizeof( m_node_v);
			swap_stat_v.n_pair_v --;
			//s

			//Update the max_d property.
			update_max_d( &context_v->m_list[p_id]);
		}
		else										//The m_node remained and the weight is substracted.
		{
			context_v->m_list[p_id].next->wei -= wei;
		}
	}
	else				//Case 2: similar process to that of Case 1.
	{
		//t
		if( m_node_v->next->wei < wei)		//This case NEVER happens.
		{
			fprintf( stderr, "Attempt to break down the matching pair with more than its weight.\n");
			exit( 0);
		}
		//t

		if( m_node_v->next->wei == wei)		//The m_node should be deleted.
		{
			m_node_v1 = m_node_v -> next;
			m_node_v->next = m_node_v->next->next;
			free( m_node_v1);

			//s
			swap_stat_v.memory_v -= sizeof( m_node_v);
			swap_stat_v.n_pair_v --;
			//s

			//Update the max_d property.
			update_max_d( &context_v->m_list[p_id]);
		}
		else								//The m_node remained and the weight is substracted.
		{
			m_node_v->next->wei -= wei;
		}
	}
}
*/

/*
 *	The new implementation of break_match_v2.
 *	
 *	Added by Cheng on 27 Feb.
 */
void break_match( context_t* context_v, int p_id, int o_id, FLOAT wei)
{
	m_node_t* m_node_v1, *m_node_v2;

	//Locate the target match.
	m_node_v1 = context_v->m_list[ p_id].head;
	m_node_v2 = m_node_v1->next;

	while( m_node_v2 != NULL && m_node_v2->id < o_id)
	{
		m_node_v1 = m_node_v2;
		m_node_v2 = m_node_v1->next;

		if( m_node_v2->id > o_id)
			break;
	}

	//t
	if( m_node_v2 == NULL || m_node_v2->id > o_id)
	{
		fprintf( stderr, "[break_match] Cannot find the target match.\n");
		exit( 0);
	}
	//t

	//m_node_v2 corresponds to the target match.
	//Break the match.
	m_node_v2->wei -= wei;

	//t
	if( m_node_v2->wei < 0)
	{
		fprintf( stderr, "[break_match] The weight inconsistency of the target match.\n");
		exit( 0);
	}
	//t
		
	if( m_node_v2->wei == 0)
	{
		//The match should be deleted from the list.
		//And some corresponding information should be updated.
		m_node_v1->next = m_node_v2->next;

		if( context_v->m_list[ p_id].rear == m_node_v2)
			context_v->m_list[ p_id].rear = m_node_v1;

		if( context_v->m_list[ p_id].max_d = m_node_v2)
		{
			//
			update_max_d( context_v, p_id);
		}

		free( m_node_v2);

		//s
		swap_stat_v.n_pair_v --;
		swap_stat_v.memory_v -= sizeof( m_node_t);
		//s
	}

	//Update the list.
	context_v->m_list[ p_id].avai += wei;
}

/*
* swap_operation performs the "swapping operation" along a swapping chain.
*
* @context_v indicates the data info.
* @sc indicates the "swapping chain", along which the "swapping operation" is performed.
*/
void swap_operation( context_t* context_v, s_chain* sc)
{
	int i;
	for( i=0; i<sc->leng-2; i+=2)
	{
		break_match( context_v, sc->chain[ i+1], sc->chain[ i+2], sc->amount);

		form_match( context_v, sc->chain[ i+1], sc->chain[ i], sc->amount, sc->dist[ i+1]);
	}

	form_match( context_v, sc->chain[ i+1], sc->chain[ i], sc->amount, sc->dist[ i+1]);
}

/*
* Locate_mmd locates the matching pair with the distance being mmd value.
*
* @context_v indicates the context info.
* @o_id store the customer id in the targetted matching pair.
*
* @Return the current mmd value.
*
DOUBLE Locate_mmd( context_t* context_v, int* o_id)
{	
	int i, p_id;
	DOUBLE mmd;

	//Locate the matching pair with the matching distance being "mmd" value.
	mmd = -1;
	for( i=0; i<context_v->p_objs_num; i++)
	{
		if( context_v->m_list[i].max_d != NULL && context_v->m_list[i].max_d->dist > mmd)
		{
			p_id = i;
			mmd = context_v->m_list[i].max_d->dist;
		}
	}

	//Updated on 27 Feb. 2013.
	m_node_v1 = context_v->m_list[ p_id].next;
	while( m_node_v1 != NULL)
	{
		if( m_node_v1->dist == mmd)
			break;

		m_node_v1 = m_node_v1->next;
	}

	//t
	if( m_node_v1 == NULL)
	{
		fprintf( stderr, "The matching pair with mmd value cannot be found.\n");
		exit(0);
	}
	//t

	//t
	if( mmd < 0)
	{
		fprintf( stderr, "[Locate_mmd] The mmd cannot be located.\n");
		exit( 0);
	}
	//t
	
	// *o_id = m_node_v1->id;
	*o_id = context_v->m_list[ p_id].max_d->id;

	return mmd;
}*/

/*
* break_match_mmd breaks the matching pair with the matching distance of "mmd" value.
*
* @context_v indicates the context information.
*/
void break_match_mmd( context_t* context_v)
{	
	int i, p_id;
	DOUBLE mmd;
	m_node_t* m_node_v1, *m_node_v2;

	//Updated on 17 Feb. 2013.
	//Locate the matching pair with the matching distance being "mmd" value.
	p_id = 0;
	mmd = -1;
	for( i=0; i<context_v->p_objs_num; i++)
	{
		if( context_v->m_list[i].max_d != NULL && context_v->m_list[i].max_d->dist > mmd)
		{
			p_id = i;
			mmd = context_v->m_list[ i].max_d->dist;
		}
	}

	//t
	if( mmd < 0)
	{
		fprintf( stderr, "[break_match_mmd] Cannot find the mmd.\n");
		exit( 0);
	}
	//t

	context_v->mmd = mmd;
	context_v->p_id = p_id;
	context_v->o_m = context_v->m_list[ p_id].max_d->id;

	//Locate the node just before the mmd_node.
	//p_id = context_v->p_id;

	m_node_v1 = context_v->m_list[ p_id].head;
	while( m_node_v1->next != context_v->m_list[ p_id].max_d)
		m_node_v1 = m_node_v1->next;

//t
	if( m_node_v1->next != context_v->m_list[ p_id].max_d)
	{
		fprintf( stderr, "[break_match_mmd] mmd inconsistency.\n");
		exit(0);
	}
//t

	//Delete mmd match.
	m_node_v2 = m_node_v1->next;
	m_node_v1->next = m_node_v2->next;

	context_v->m_list[ p_id].avai += m_node_v2->wei;
	update_max_d( context_v, p_id);
	if( context_v->m_list[ p_id].rear == m_node_v2)
		context_v->m_list[ p_id].rear = m_node_v1;

	context_v->o_m_dem = m_node_v2->wei;

	free( m_node_v2);

	//s
	swap_stat_v.n_pair_v --;
	swap_stat_v.memory_v -= sizeof( m_node_t);
	//s
}

/*
* Update_Graph_List updates the g_list array w.r.t. the new mmd value.
*
* @context_v indicates the context data.
* @mmd indicates the new mmd value.
*
* @Return 1 when updated distance list is normal; otherwise return 0.
*
int Update_Graph_List( context_t* context_v, DOUBLE mmd)
{
	int i, n_d_edge;
	g_node_t* g_node_v1, *g_node_v2, *g_node_v3;

	n_d_edge = 0;
	for( i=0; i<context_v->o_objs_num; i++)
	{
		g_node_v1 = context_v->g_list[i].next;
		
		//t
		if( g_node_v1 == NULL)
		{
			fprintf( stderr, "No P objects are within the distance of mmd from \"o_id=%i\".\n", i);
			exit(0);
		}
		//t	

		//Deal with the boundary case--deleting the first part of the list.
		while( g_node_v1 != NULL && g_node_v1->dist > mmd)
		{
			g_node_v2 = g_node_v1;
			g_node_v1 = g_node_v1->next;
			context_v->g_list[i].next = g_node_v1;

			free( g_node_v2);

			//s
			swap_stat_v.memory_v -= sizeof( g_node_t);
			n_d_edge ++;
			//s
		}
		
		if( g_node_v1 == NULL)
		{
			//s
			swap_stat_v.n_edge_sum += swap_stat_v.n_prev_edge - n_d_edge;
			swap_stat_v.n_prev_edge -= n_d_edge;
			//s

			return 0;
		}

		//Deal with the normal case.
		g_node_v2 = g_node_v1->next;
		while( g_node_v2 != NULL)
		{
			if( g_node_v2->dist > mmd)
			{
				g_node_v3 = g_node_v2;
				g_node_v2 = g_node_v2->next;
				g_node_v1->next = g_node_v2;

				free( g_node_v3);

				//s
				swap_stat_v.memory_v -= sizeof( g_node_t);
				n_d_edge ++;
				//s

				continue;
			}

			g_node_v1 = g_node_v2;
			g_node_v2 = g_node_v2->next;
		}
	}//for(i);

	//s
	swap_stat_v.n_edge_sum += swap_stat_v.n_prev_edge - n_d_edge;
	swap_stat_v.n_prev_edge -= n_d_edge;
	//s

	return 1;
}
*/

/*
* Swapping_Chain performs the "Swapping_Chain" algorithm on a full-matching involved in context_v.
*
* @context_v indicates the input data, whose matching variable has been initialized into a full-matching.
*/
DOUBLE Swapping_Chain( context_t* context_v)
{
	float n_swap_v;
	s_chain* sc;

	int cnt = 0;

	int	interval = 3000;

	//The initialization work is done outside the function.

	sc = NULL;
	while( true)
	{
		//s
		n_swap_v = 0;
		//s

		//Break down the matching pair (o_m, p_m) with the maximum matching distance.
		break_match_mmd( context_v);

#if		DISCRITIZE_TAG
		context_v->mmd = ( float)floor( context_v->mmd) - interval;

		//if( interval > 5)
		//	interval -= 5;
		//else interval = 1;
#endif

		//t
		cnt ++;
		if( cnt % 100 == 0)
			printf( "mmd: %f\n", context_v->mmd);
		//t
	
		while( context_v->o_m_dem != 0)
		{
			if( graph_impl_tag == 1)
				sc = search_sc_graph( context_v);
			else
				sc = search_sc_tree( context_v);

			if( sc == NULL)
				break;

			//s
			n_swap_v ++;
			//s

			swap_operation( context_v, sc);
			context_v->o_m_dem -= sc->amount;

			//s
			swap_stat_v.memory_v -= sizeof( s_chain_t) + 
				sizeof( int) * sc->leng + sizeof( DOUBLE) * sc->leng;
			//s

			free( sc->chain);
			free( sc->dist);
			free( sc);

			sc = NULL;
		}

		if( context_v->o_m_dem == 0)		//o_m is re-satisfied.
		{
			/*
			if( sc != NULL)
			{
				//s
				swap_stat_v.memory_v -= sizeof( s_chain_t) + 
					sizeof(int) * sc->leng + sizeof(DOUBLE) *sc->leng;
				//s

				free( sc->chain);
				free( sc->dist);
				free( sc);
				sc = NULL;
			}
			*/

			//s
			swap_stat_v.n_reduction ++;
			if( n_swap_v < swap_stat_v.n_swap.min)
				swap_stat_v.n_swap.min = n_swap_v;
			if( n_swap_v > swap_stat_v.n_swap.max)
				swap_stat_v.n_swap.max = n_swap_v;
			swap_stat_v.n_swap_sum += n_swap_v;
			//s

			//Updated on 27 Feb. 2013.
			//Update the information of "mmd", "o_m".
			//context_v->mmd = Locate_mmd( context_v, &context_v->o_m);	//&context_v->p_m

			//Update the graph list according to the new mmd value.
			/*
			if( !Update_Graph_List( context_v))	//Why update? Early termination.
				return context_v->mmd;
			*/

			continue;
		}
		else						//o_m cannot be satisfied.
		{
			return context_v->mmd;
		}
	}
}

/*
* Test_Swapping_Chain tests the implementation of the "Swapping_Chain" algorithm.
* @opt indicates the choices of "ini_context".
*	1 for Non-Heuristic method.(random)
*	2 for Heuristic method.(Sorting & Partially random)

void Test_Swapping_Chain( int opt)
{
	DOUBLE mmd;

	d_config_t* cfg;
	data_t* data_v;
	context_t* context_v;

	FILE* o_fp;

	//Read the configuration info.
	cfg = dReadConfig( );

	//Read the raw data.
	data_v = dReadData( cfg);

	//Initialize the the statistic information.
	memset( &swap_stat_v, 0, sizeof( stat_t));

#ifndef WIN32
	float ini_u_time, ini_s_time, match_u_time, match_s_time;
	struct rusage ini_start, ini_end, match_end;
	FILE* s_fp;
	
	get_cur_time( &ini_start);
#endif

	//Initialize the context data
	if( opt == 1)
		context_v = random_ini_context( );
	else if( opt == 2)
		context_v = sort_ini_context( );

#ifndef WIN32 
	get_cur_time( &ini_end);
#endif

	//Perform the fair-MM calculation.
	mmd = Swapping_Chain( context_v);

#ifndef WIN32
	get_cur_time( &match_end);
	
	collect_time( &ini_start, &ini_end, &ini_u_time, &ini_s_time);
	collect_time( &ini_end, &match_end, &match_u_time, &match_s_time);

	if( (s_fp = fopen( cfg->stat_file, "w")) == NULL)
	{
		fprintf( stderr, "The statistic file cannot be open.\n");
		exit(0);
	}

	fprintf( s_fp, "Initialization user_time: %f\t system_time: %f\n", ini_u_time, ini_s_time);
	fprintf( s_fp, "Matching user_time: %f\t system_time: %f\n", match_u_time, match_s_time);
	fprintf( s_fp, "Total user_time: %f\t system_time: %f\n", ini_u_time + match_u_time, ini_s_time + match_s_time);

	fclose( s_fp);
#endif

	//Print the results.
	if( (o_fp = fopen( cfg->matching, "w")) == NULL)
	{
		fprintf( o_fp, "The matching result file cannot be open.\n");
		exit(0);
	}

	fprintf( o_fp, "The mmd value: %lf\n", mmd);

	//Release the related the resources.
	free( cfg);
	FreeData( );
	context_release( context_v);

	fclose( o_fp);
	
}*/

/*
* emp_Swapping_Chain tests the implementation of the "Swapping_Chain" algorithm.
* @opt indicates the choices of "ini_context".
*	1 for Non-Heuristic method.(random)
*	2 for Heuristic method.(Sorting & Partially random)
*	3 for kNN method.
*	4 for fair method.
*/
void emp_Swapping_Chain( )
{
	DOUBLE mmd;

	_config_t* cfg;
	//data_t* data_v;
	context_t* context_v;

	//Initialize the the statistic information.
	memset( &swap_stat_v, 0, sizeof( stat_t));

	swap_stat_v.n_swap.min = FLT_MAX;
	swap_stat_v.n_swap_que.min = FLT_MAX;
	swap_stat_v.n_expension.min = FLT_MAX;
//	swap_stat_v.n_edge.min = INT_MAX;
//	swap_stat_v.n_pair.min = INT_MAX;

	//
	context_v = context_alloc( );

	//Read the configuration info.
	//printf( "Read configuration...\n");
	cfg = emp_read_config( );

	graph_impl_tag = cfg->graph_tag;

	//Read the raw data.
	printf( "Reading data ...\n");
	context_v->data_v = emp_read_data( cfg);

	//s
	swap_stat_v.memory_v += sizeof( data_t) + 
							sizeof( obj_t) * ( context_v->data_v->n_p_objs + context_v->data_v->n_o_objs) +
							sizeof( FLOAT) * context_v->data_v->dim * ( context_v->data_v->n_p_objs + context_v->data_v->n_o_objs);
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

#ifndef WIN32
	struct rusage r_tree_start;
	get_cur_time( &r_tree_start);
#endif

	if( graph_impl_tag == 0)
	{
		//Build the R-tree on P.
		printf( "Building R-tree on P ...\n");

		build_R_tree( context_v->data_v);
	}

#ifndef WIN32
	struct rusage ini_start, ini_end, match_end;
	
	get_cur_time( &ini_start);
#endif

	//Initialize the context data.
	context_alloc_part( context_v);

//	if( cfg->swap_opt == 1)
//	{
//		printf( "Initializing a full assignment (Swap-Random) ...\n");
//		random_ini_match_list( context_v);

		//random_ini_context( context_v);
//	}
//	else 
	if( cfg->swap_opt == 1)
	{
		printf( "Initializing a full assignment (Swap-Sort) ...\n");
		sort_ini_match_list( context_v);
	}
//	else if( cfg->swap_opt ==3)
//	{
//		printf( "Initializing a full assignment (Swap-kNN) ...\n");
//		kNN_ini_match_list( context_v);
//	}
	else
	{
		printf( "Initialize a full assignment (Swap-Fair) ...\n");
		fair_ini_match_list( context_v);
	}

	ini_context_member( context_v);
	if( graph_impl_tag == 1)
		ini_graph_list( context_v, context_v->mmd);

#ifndef WIN32 
	get_cur_time( &ini_end);
#endif

	//Perform the fair-MM calculation.
	printf( "Swap-Chain ...\n");
	mmd = Swapping_Chain( context_v);

#ifndef WIN32
	float r_tree_build_time, tmp, ini_u_time, ini_s_time, match_u_time, match_s_time;

	get_cur_time( &match_end);
	
	collect_time( &r_tree_start, &ini_start, &r_tree_build_time, &tmp);
	collect_time( &ini_start, &ini_end, &ini_u_time, &ini_s_time);
	collect_time( &ini_end, &match_end, &match_u_time, &match_s_time);

	swap_stat_v.preprocess_time = ini_u_time;
	swap_stat_v.matching_time = match_u_time;
	swap_stat_v.total_time = ini_u_time + match_u_time;

	if( cfg->thr_opt == -1 && graph_impl_tag == 0 && cfg->swap_opt != 4)
	{
		swap_stat_v.total_time += r_tree_build_time;
	}

#endif

	//Calculate the related statistics.
	if( swap_stat_v.n_swap.min == FLT_MAX)
		swap_stat_v.n_swap.min = 0;
	if( swap_stat_v.n_swap_que.min == FLT_MAX)
		swap_stat_v.n_swap_que.min = 0;
	if( swap_stat_v.n_expension.min == FLT_MAX)
		swap_stat_v.n_expension.min = 0;
	
	if( swap_stat_v.n_reduction != 0)
		swap_stat_v.n_swap.aver = swap_stat_v.n_swap_sum/swap_stat_v.n_reduction;
	if( swap_stat_v.n_swap_sum != 0)
		swap_stat_v.n_swap_que.aver = swap_stat_v.n_swap_que_sum/swap_stat_v.n_swap_sum;
	if( swap_stat_v.n_swap_sum !=0)
		swap_stat_v.n_expension.aver = swap_stat_v.n_expension_sum/swap_stat_v.n_swap_sum;
	if( swap_stat_v.n_reduction != 0)
		swap_stat_v.n_edge_aver = swap_stat_v.n_edge_sum/swap_stat_v.n_reduction;

	/*
#ifndef WIN32
	if( cfg->swap_opt == 3)
	{ 
		float	u_t_r_tree, s_t_r_tree;
		collect_time( &ini_start, &t_rTree_end, &u_t_r_tree, &s_t_r_tree);
		fprintf( s_fp, "R_tree Construction: %f\n", u_t_r_tree);
	}
#endif
	*/

	//s
	swap_stat_v.memory_v -= sizeof( data_t) + 
							sizeof( obj_t) * ( context_v->data_v->n_p_objs + context_v->data_v->n_o_objs) +
							sizeof( FLOAT) * context_v->data_v->dim * ( context_v->data_v->n_p_objs + context_v->data_v->n_o_objs);
	//s

	//Print the statistics.
	print_stat( mmd, swap_stat_v);

	//Release the related the resources.
	free( cfg);
	//FreeData( );
	context_release( context_v);

	if( graph_impl_tag == 0)
		free_R_tree( );

	printf( "mmd: %lf\n", mmd);
	printf( "total time: %f\n", swap_stat_v.total_time);
//	printf( "R-tree building time for Fair-Swap: %f\n", swap_stat_v.R_tree_build_time);
//	printf( "Memory balance: %f\n\n", swap_stat_v.memory_v);
	printf( "Memory (MB): %f\n", swap_stat_v.memory_max / ( 1024 * 1024));
}



//Updated on 27 Feb.
//---------------------------------------

/*
 *	Allocate a p_list_t structure.
 */
p_list_t* p_list_alloc( )
{
	p_list_t* p_list_v;

	p_list_v = ( p_list_t*)malloc( sizeof( p_list_t));
	memset( p_list_v, 0, sizeof( p_list_t));

	p_list_v->p_head = ( p_node_t*)malloc( sizeof( p_node_t));
	memset( p_list_v->p_head, 0, sizeof( p_node_t));


	//s
	swap_stat_v.memory_v += sizeof( p_list_t) + sizeof( p_node_t);
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	return p_list_v;
}

/*
 *
 */
void add_p_list_entry( p_list_t* p_list_v, int p_id, DOUBLE dist)
{
	p_node_t* p_node_v;

	p_node_v = ( p_node_t*)malloc( sizeof( p_node_t));
	memset( p_node_v, 0, sizeof( p_node_t));

	p_node_v->p_id = p_id;
	p_node_v->dist = dist;

	p_node_v->next = p_list_v->p_head->next;
	p_list_v->p_head->next = p_node_v;

	p_list_v->num ++;


	//s
	swap_stat_v.memory_v += sizeof( p_node_t);
	if( swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	return;
}

/*
 *	Release a p_list_t structure.
 */
void p_list_release( p_list_t* p_list_v)
{
	p_node_t* p_node_v1, *p_node_v2;

	p_node_v1 = p_list_v->p_head;
	while( p_node_v1 != NULL)
	{
		p_node_v2 = p_node_v1->next;
		free( p_node_v1);

		p_node_v1 = p_node_v2;

		//s
		swap_stat_v.memory_v -= sizeof( p_node_t);
		//s
	}

	free( p_list_v);

	//s
	swap_stat_v.memory_v -= sizeof( p_list_t);
	//s
}

/*
 *	Calculate the minimum distance between the MBR of a node @node_v 
 *	and a location @loc_v.
 */
DOUBLE calc_minDist( range_t* MBR, obj_t* obj_v)
{
	int i;
	DOUBLE m_dist;
	
	//Calculate the minimum distance between the MBR and the location.
	m_dist = 0;
	for( i=0; i<R_tree_v.dim; i++)
	{
		if( obj_v->axis[ i] < MBR[ i].min)
			m_dist += pow( MBR[ i].min - obj_v->axis[ i], 2);
		else if( obj_v->axis[ i] > MBR[ i].max)
			m_dist += pow( obj_v->axis[ i] - MBR[ i].max, 2);
	}

	return ( DOUBLE)sqrt( m_dist);
}

/*
 *	Calculate the maxDist between a MBR @MBR and 
 */
DOUBLE calc_maxDist( range_t* MBR, obj_t* obj_v)
{
	int i;
	DOUBLE maxDist;

	maxDist = 0;
	for( i=0; i<R_tree_v.dim; i++)
	{
		if( obj_v->axis[ i] <= ( MBR[ i].min + MBR[ i].max) / 2)
			maxDist += pow( MBR[ i].max - obj_v->axis[ i], 2);
		else
			maxDist += pow( obj_v->axis[ i] - MBR[ i].min, 2);
	}

	return sqrt( maxDist);
}

/*
 *	Calculate the minimum distance between the MBR of a node @node_v and a location @loc_v.
 */
DOUBLE calc_minDist_node( tree_node_t* node_v, obj_t* obj_v)
{
	DOUBLE dist;
	range_t* MBR;

	if( node_v->parent == NULL)
		MBR = get_MBR_node( node_v, R_tree_v.dim);
	else
		MBR = node_v->parent->MBRs[ node_v->loc];

	dist = calc_minDist( MBR, obj_v);

	if( node_v->parent == NULL)
	{
		free( MBR);

		//s
		swap_stat_v.memory_v -= sizeof( R_tree_v.dim * sizeof( range_t));
		//s
	}

	return dist;
}

/*
 *	Check whether a MBR @MBR overlaps with a disk @disk_v.
 */
bool is_overlap( range_t* MBR, obj_t* obj_v, DOUBLE radius)
{
	DOUBLE min_dist;

	min_dist = calc_minDist( MBR, obj_v);
	if(  min_dist < radius - DELTA)
		return true;
	else
		return false;
}

/*
 *	Check whether a MBR @MBR is enclosed entirely by a disk @disk_v.
 */
bool is_enclosed( range_t* MBR, obj_t* obj_v, DOUBLE radius)
{
	DOUBLE dist;

	dist = calc_maxDist( MBR, obj_v);

	if( dist < radius - DELTA)
		return true;
	else
		return false;
}

/*
 *	Retrieve all the objects located at the sub-tree rooted at @node_v.
 *	The retrieved objects are stored in obj_set_v.
 */
void retrieve_sub_tree( tree_node_t* node_v, p_list_t* p_list_v, obj_t* obj_v)
{
	int i;
	obj_t* obj_temp;
	DOUBLE dist;

	if( node_v->level == 0)
	{
		//node_v is a leaf-node.
		//Retrieve all its objects.
		for( i=0; i<node_v->num; i++)
		{
			obj_temp = ( obj_t*)( node_v->child[ i]);
			dist = calc_dist( obj_temp, obj_v, R_tree_v.dim);
			add_p_list_entry( p_list_v, obj_temp->id, dist);
		}
	}
	else
	{
		//node_v is an inner-node.
		//Invoke the process recursively.
		for( i=0; i<node_v->num; i++)
		{		
			retrieve_sub_tree( ( tree_node_t*)( node_v->child[ i]), p_list_v, obj_v);
		}
	}
}

/*
 *	Range query on the sub-tree rooted at @node_v.
 *	@disk_v indicates the range_t which is a circle.
 *
 *	The results are stored in @obj_set_v.
 */
void range_query_sub( tree_node_t* node_v, obj_t* obj_v, DOUBLE radius, p_list_t* p_list_v)
{
	int i;
	obj_t* obj_tmp;
	DOUBLE dist;
	range_t* MBR;

	if( node_v->parent == NULL)
		MBR = get_MBR_node( node_v, R_tree_v.dim);
	else
		MBR = node_v->parent->MBRs[ node_v->loc];

	if( !is_overlap( MBR, obj_v, radius))	//No overlapping.
	{
		if( node_v->parent == NULL)
		{
			free( MBR);

			//s
			swap_stat_v.memory_v -= R_tree_v.dim * sizeof( range_t);
			//s
		}

		return;
	}

	if( is_enclosed( MBR, obj_v, radius))	//Enclosed entrely.
	{
		retrieve_sub_tree( node_v, p_list_v, obj_v);

		if( node_v->parent == NULL)
		{
			free( MBR);

			//s
			swap_stat_v.memory_v -= R_tree_v.dim * sizeof( range_t);
			//s
		}

		return;
	}

	//The remaining cases.
	if( node_v->level == 0)
	{
		//node_v is a leaf-node.
		for( i=0; i<node_v->num; i++)
		{
			obj_tmp = ( obj_t*)( node_v->child[ i]);
			if( is_enclosed( obj_tmp->MBR, obj_v, radius))
			{
				dist = calc_dist( obj_tmp, obj_v, R_tree_v.dim);
				add_p_list_entry( p_list_v, obj_tmp->id, dist);
			}
		}
	}
	else
	{
		//node_v is an inner-node.		
		for( i=0; i<node_v->num; i++)
		{
			range_query_sub( ( tree_node_t*)( node_v->child[ i]), obj_v, radius, p_list_v);
		}
	}

	if( node_v->parent == NULL)
	{
		free( MBR);

		//s
		swap_stat_v.memory_v -= R_tree_v.dim * sizeof( range_t);
		//s
	}
}

/*
 *	Circle range_t query.
 *
 *	DFS: recursive implementation.
 */
p_list_t* range_query( obj_t* obj_v, DOUBLE radius)
{
	p_list_t* p_list_v;

	p_list_v = p_list_alloc( );

	range_query_sub( R_tree_v.root, obj_v, radius, p_list_v);

	return p_list_v;
}

/*
 *	The implementation of search_sc_graph based on R-tree.
 */
s_chain* search_sc_tree( context_t* context_v)
{
	int cur, rear = 0, suc = 0, o_id;
	int* p_tag, *o_tag;
	float expension_v = 0;

	q_node_t* q_node_v, *q_node_tmp;
	m_node_t* m_node_v;

	p_list_t* p_list_v;
	p_node_t* p_node_v;

	q_node_t** que;
	s_chain_t* sc;

	//Allocate the main memory.
	que = ( q_node_t**)malloc( sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num));
	memset( que, 0, sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num));

	p_tag = ( int*)malloc( sizeof( int) * context_v->p_objs_num);
	o_tag = ( int*)malloc( sizeof( int) * context_v->o_objs_num);
	memset( p_tag, 0, sizeof( int) * context_v->p_objs_num);
	memset( o_tag, 0, sizeof( int) * context_v->o_objs_num);
	
	//s
	swap_stat_v.memory_v += sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num) +
							sizeof( int) * context_v->p_objs_num + sizeof( int) * context_v->o_objs_num;
	//s

	//Push o_m into the queue.
	q_node_v = ( q_node_t*)malloc( sizeof( q_node_t));
	memset( q_node_v, 0, sizeof( q_node_t));

	q_node_v->id = context_v->o_m;
	q_node_v->tag = 0;
	q_node_v->amount = context_v->o_m_dem;
	q_node_v->dist = 0;
	q_node_v->prev = -1;
	o_tag[ context_v->o_m] = 1;

	que[ rear++] = q_node_v;
	cur = 0;

	//Searching the "swapping chain" using BFS method.
	while( cur != rear)
	{
		q_node_tmp = que[ cur];
		if( q_node_tmp->tag == 0)		//The current node is a O object.
		{
			//Push all the P objects that are within the distance of mmd from the O object being delt with currently.
			//Perform a range_t query here.
			o_id = q_node_tmp->id;
			p_list_v = range_query( &context_v->data_v->o_objs[ o_id], context_v->mmd);

			p_node_v = p_list_v->p_head->next;
			while( p_node_v != NULL)
			{
				//s
				expension_v ++;
				//s

				if( p_tag[ p_node_v->p_id] == 0)
				{
					q_node_v = (q_node_t*)malloc( sizeof( q_node_t));
					memset( q_node_v, 0, sizeof( q_node_t));

					q_node_v->id = p_node_v->p_id;
					q_node_v->dist = p_node_v->dist;
					q_node_v->tag = 1;
					q_node_v->amount = q_node_tmp->amount;	//temporary use. INF is OK here.
					q_node_v->prev = cur;

					p_tag[ p_node_v->p_id] = 1;
					
					que[ rear++] = q_node_v;

					if( context_v->m_list[ p_node_v->p_id].avai > 0)	//Forming a swapping chain, stop.
					{						
						suc = 1;
						goto end;
					}
				}

				p_node_v = p_node_v->next;
			}//while

			p_list_release( p_list_v);
		}
		else							//The current node is a P object.
		{
			//Push all the O objects that are currently matched with the P object being dealt with.
			m_node_v = context_v->m_list[ q_node_tmp->id].head->next;
			while( m_node_v != NULL)
			{
				//s
				expension_v ++;
				//s

				if( o_tag[ m_node_v->id] == 0)	//un-tagged O object.
				{
					q_node_v = (q_node_t*)malloc( sizeof( q_node_t));
					memset( q_node_v, 0, sizeof( q_node_t));

					q_node_v->id = m_node_v->id;
					//q_node_v->dist = 0;		
					q_node_v->tag = 0;
					q_node_v->amount = m_node_v->wei;
					q_node_v->prev = cur;

					o_tag[ m_node_v->id] = 1;

					que[ rear++] = q_node_v;
				}
				
				m_node_v = m_node_v->next;
		
			}//while
		}//else

		cur ++;
	}//while(cur != rear)

end:
	//s
	swap_stat_v.memory_v += sizeof( q_node_t) * rear;

	if( rear < swap_stat_v.n_swap_que.min)
		swap_stat_v.n_swap_que.min = ( float)rear;
	if( rear > swap_stat_v.n_swap_que.max)
		swap_stat_v.n_swap_que.max = ( float)rear;
	swap_stat_v.n_swap_que_sum += ( float)rear;

	if( expension_v > swap_stat_v.n_expension.max)
		swap_stat_v.n_expension.max = expension_v;
	if( expension_v < swap_stat_v.n_expension.min)
		swap_stat_v.n_expension.min = expension_v;
	swap_stat_v.n_expension_sum += ( float)expension_v;
	//s

	if( suc == 1)	//Case 1: A swapping chain has been found.
		sc = construct_sc( que, rear);
	else			//Case 2: No swapping chain exists.
	{
		sc = NULL;
	}

	//s
	if( swap_stat_v.memory_v > swap_stat_v.memory_max)
		swap_stat_v.memory_max = swap_stat_v.memory_v;
	//s

	//Release the related the resources.
	cur = 0;
	while( cur < rear)
		free( que[ cur++]);
	free( que);
	free( o_tag);
	free( p_tag);

	//s
	swap_stat_v.memory_v -= sizeof( q_node_t) * rear;
	swap_stat_v.memory_v -= sizeof( q_node_t*) * (context_v->p_objs_num + context_v->o_objs_num) +
							sizeof( int) * context_v->p_objs_num + sizeof( int) * context_v->o_objs_num;
	//s
	
	return sc;
}