
/*
* Swap_Chain.h contains the declarations related the implementation of acontext_vorithm Swapping-Chain.
*
* 1. We store the network graph in form of "adjacent list",		--Un-ordered.
*	with the head being a customer and nodes being service providers linked to the head.
* 
* 2. We also store the matching data in form of "list array",	--Ordered.
*	with the head being a service provider and nodes being customers matched to the head.
*
* 3. The id of nodes begin with 0.
*
* 4. The cost of Locating.
*/

#ifndef	SWAP_CHAIN_H
#define SWAP_CHAIN_H

#include <float.h>
//#include "dataStruct.h"
#include "dataUtility.h"
//#include "kNN.h"
#include "SPM.h"


#define DELTA 0.000001


#define	SWAP_RESULT_FILE		"mmd.txt"
#define	SWAP_STATISTIC_FILE		"stat.txt"
#define	SPM_RESULT_FILE			"fair_mmd.txt"

#define	EMP_NORM_CONFIG_FILE	"norm_config.txt"
#define EMP_REAL_GEN_CONFIG		"real_config.txt"


#define	DISCRITIZE_TAG	0


extern	int	graph_impl_tag;	//Indicating whether maintaining a graph for searching swapping-chains.


//The structure of the node for storing the graph data.
typedef struct g_node			
{
	int				id;		//the id of a node.
	DOUBLE			dist;	//the distance from the head.
	struct g_node*	next;	//

}	g_node_t;

//The structure of the head node of graph list.
typedef struct g_head
{
//	DOUBLE		max_d;		//the maximum distance among all the distances from the P objects to the head(O object).
	g_node_t*	head;		

}	g_head_t;

//The structure of the node for storing the matching data.
typedef struct m_node			
{
	int				id;		//the id of a node.
	FLOAT			wei;	//the matching weight.
	DOUBLE			dist;	//the matching distance.

	struct m_node*	next;

}	m_node_t;

//The structure of the head node of the matching list.
typedef struct m_head			
{
	FLOAT		cap;	//the initial service capacity of a service provider.
	FLOAT		avai;	//the available service capacity of a service provider.
/*
	DOUBLE		max_d;	//the maximum matching distance of the head(P object).
	int			id;		//the id of the O object owning the maximum matching distance from the head(P object).
*/
	m_node_t*   max_d;	//the pointer pointing the O object owning the maximum matching distance from the head(P object).	

	m_node_t*	head;
	m_node_t*	rear;

}	m_head_t;
	
//The structure of the node in the queue used for searching the "swapping chain".
typedef struct q_node
{
	int		id;				//the id of the node.
	int		tag;			//the tag of the node.
								//0 for O objects.
								//2 for P objects.
	DOUBLE	dist;			//the matching distance info.
	FLOAT	amount;			//the information used to calculate the "swapping amount" info.
	int		prev;			//the identifier of the former node.

}	q_node_t;

//The structure of the swapping chain.
typedef struct s_chain
{
	int*	chain;			//the array representing the chain.
	DOUBLE* dist;			//the array representing the distance info w.r.t. the chain.

	int		leng;			//the length of the chain.
	FLOAT	amount;			//the corresponding the "swapping amount".

}	s_chain_t;

//The structure storing the dist info for sorting.
typedef struct p_dist
{
	int		p_id;
	DOUBLE	dist;
	FLOAT	avai;

}	p_dist_t;

//The structure of the data used by the Swapping_Chain acontext_vorithm.
typedef struct context			
{
	int			p_objs_num;	//the num of P objects.
	int			o_objs_num;	//the num of O objects.

	int			o_m;		//the current O object that is matched with the distance of mmd value.
//	int			p_m;		//the current P object that is matched with the distance of mmd value.
	FLOAT		o_m_dem;	//the service demand of o_m.

	DOUBLE		mmd;		//the current maximum matching distance.
	int			p_id;		//the product involved in the mmd match.
//	m_node_t*	m_node_v_mmd;
	
	//int*		p_tag;		//the tag information of P objects.
	//int*		o_tag;		//the tag information of O objects.

	g_head_t*	g_list;		//the graph information maintained in form of "adjacent list".
	m_head_t*	m_list;		//the matching information maintained in form of "list array".

	//Facilitaing variables.
	//q_node_t**  que;		//the queue used for searching for the corresponding swapping chain.


	//
	data_t*		data_v;
	
}	context_t;



//Updated on 28 Feb. 2013.
//---------------------------------------
typedef struct p_node
{
	int				p_id;
	DOUBLE			dist;

	struct	p_node*	next;

}	p_node_t;

//The structure for storing the range_t query results.
typedef struct p_list
{
	int			num;
	p_node_t*	p_head;

}	p_list_t;


/*
#ifndef WIN32

extern struct rusage t_rTree_end;	//Only used in "kNN_ini_match_list"--Construction of r_tree.

extern struct rusage t_rTree_end;

#endif
*/

extern stat_t swap_stat_v;

//---------------------------------------

inline int cmp_dist( const void* fir, const void* sec);

DOUBLE calc_dist( obj_t* obj1, obj_t* obj2, int dim);

//---------------------------------------

context_t* context_alloc( );

void context_alloc_part( context_t* context_v);

void random_ini_match_list( context_t* context_v);

void sort_ini_match_list( context_t* context_v);

void kNN_ini_match_list( context_t* context_v);

void fair_ini_match_list( context_t* context_v);

void ini_context_member( context_t* context_v);

void ini_graph_list( context_t* context_v, DOUBLE mmd);

//void random_ini_context( context_t* context_v);

//void sort_ini_context( context_t* context_v);

//void kNN_ini_context( context_t* context_v);

//void fair_ini_context( context_t* context_v);

void context_release( context_t* context_v);

//---------------------------------------

s_chain* construct_sc( q_node_t** que, int rear);

s_chain* search_sc_graph( context_t* context_v);

//---------------------------------------

//m_node_t* Locate_Match_Form( context_t* context_v, int p_id, int o_id);

//m_node_t* Locate_Match_Break( context_t* context_v, int p_id, int o_id);

void update_max_d( context_t* context_v, int p_id);

void form_match( context_t* context_v, int p_id, int o_id, FLOAT wei, DOUBLE dist);

void break_match( context_t* context_v, int p_id, int o_id, FLOAT wei);

//DOUBLE Locate_mmd( context_t* context_v, int* p_id, int* o_id);

void break_match_mmd( context_t* context_v);

//int Update_Graph_List( context_t* context_v, DOUBLE mmd);

void swap_operation( context_t* context_v, s_chain* sc);

//---------------------------------------

DOUBLE Swapping_Chain( context_t* context_v);

void emp_Swapping_Chain( );


//Updated on 27 Feb.
//---------------------------------------

p_list_t* p_list_alloc( );

void add_p_list_entry( p_list_t* p_list_v, int p_id, DOUBLE dist);

void p_list_release( p_list_t* p_list_v);

DOUBLE calc_minDist( range_t* MBR, obj_t* obj_v);

DOUBLE calc_maxDist( range_t* MBR, obj_t* obj_v);

DOUBLE calc_minDist_node( tree_node_t* node_v, obj_t* obj_v);

//obj_t* NN_key( loc_t* loc_v, KEY_TYPE key);

bool is_overlap( range_t* MBR, obj_t* obj_v, DOUBLE radius);

bool is_enclosed( range_t* MBR, obj_t* obj_v, DOUBLE radius);

//void add_obj_set_entry( obj_t* obj_v, obj_set_t* obj_set_v);

//void remove_obj_set_entry( obj_set_t* obj_set_v);

void retrieve_sub_tree( tree_node_t* node_v, p_list_t* p_list_v, obj_t* obj_v);

void range_query_sub( tree_node_t* node_v, obj_t* obj_v, DOUBLE radius, p_list_t* p_list_v);

p_list_t* range_query( obj_t* obj_v, DOUBLE radius);

s_chain* search_sc_tree( context_t* context_v);


#endif
