/*
* Threshold_Adapt.cpp is the implementations of functions in Threshold_Adapt.h.
*/

#include <string.h>
#include "Threshold_Adapt.h"


/*----------------------------Maximal Flow Algorithm-------------------------------*/


int DOUBLE_comparison(const void * a, const void * b)
{
  if ( *( DOUBLE*)a - *( DOUBLE*)b > 0)
	  return 1;
  else if ( *( DOUBLE*)a - *( DOUBLE*)b == 0)
	  return 0;
  else
	  return -1;
}

/*
* find_path_gen is used to find the augmenting path( for Algorithm 2)
* 
* @Param void.
* 
* @Return the 'O' value of queue on success, else 0.
*/
int find_path_gen( )		//Edmonds-Karp Algorithm(BFS)--Synthesize the 'normal' & 'preprocess'.
{	
	int i, j, sum = pSize + oSize, sta = 0, rear = 0;//sta
	//float expension_length = 0;

	for( i=0; i<sum; i++)
	{
		bPrev[i] = -1;				//tag.
	}
	//Put the elements in set 'P' in the bQueue.
	for( j=1; j<=pSize; j++)
	{	
		if( bGraph[j][0] - bf[j][0] > 0 )
		{
			bQueue[rear] = j;		//Store the array's order.
			bPrev[j-1] = 0;			//bug.
			bTag[rear++] = 1;		//tag=1 for P, tag=0 for O.
		}
	}//for(j)

	//BFS while(taken into situation of 'sta=rear=0').
	while( sta != rear)	
	{
		i = bQueue[sta];
		if( bTag[sta] == 1)			//P set element(Check P->O).
		{
			for( j=1; j<=oSize; j++)
			{
				if( bGraph[i][j] - bf[i][j] > 0 && bPrev[j+pSize-1] == -1)
				{
					bQueue[rear] = j;
					bPrev[ j+pSize-1] = i-1;	
					bTag[rear++] = -1;
				}
			}//for(j)

			//statistic info.
		}
		else						//O set element.
		{
			if( bGraph[0][i] - bf[0][i] > 0)
			{
				return sta;			//Not rear.
			}
			for( j=1; j<=pSize; j++)
			{
				if( bf[j][i] != 0 && bGraph[j][i] !=0 && bPrev[j-1] == -1)	//bGraph[j][i] - bf[j][i] == 0
				{
					bQueue[rear] = j;
					bPrev[ j-1] = i+pSize-1;
					bTag[rear++] = 1;
				}
			}
		}	
		sta ++;
	}//while

	/*s*/
	thr_stat_v.que_length_sum += (float)rear;
	if( rear > thr_stat_v.que_length_max)
		thr_stat_v.que_length_max = (float)rear;
	/*s*/

	return 0;
}

/*
* ford_fulkerson_gen is used to get the maximal flow for the general case.
*
* @Param/Return void.
*/
int ford_fulkerson_gen( )
{
	int flow;
	long i, j, res, cp;								//pSize = appro_var->p_data->num
	float path = 0;

	//Initialize bf array in IniGraph function.

	res = find_path_gen( );
	while( res != 0)	//(int**)graph, (int**)f, (int*&)prev
	{
		//Specify the cp value(needn't when unweighted O).
		j = bQueue[res];
		cp = bGraph[0][j] - bf[0][j];
		i = bPrev[ j+pSize-1];

		if( cp > bGraph[ i+1][j] - bf[ i+1][j])
			cp = bGraph[ i+1][j] - bf[ i+1][j];
		j = bPrev[ i];

		while( j != 0)
		{
			if( cp > bf[ i+1][ j-pSize+1])
				cp = bf[ i+1][ j-pSize+1];
			i = bPrev[j];

			if( cp > bGraph[ i+1][ j-pSize+1] - bf[ i+1][ j-pSize+1])
				cp = bGraph[ i+1][ j-pSize+1] - bf[ i+1][ j-pSize+1];
			j = bPrev[i];
		}
		if( cp > bGraph[ i+1][0] - bf[ i+1][0])
			cp = bGraph[ i+1][0] - bf[ i+1][0];
		
		//Update the augmenting path.
		j = bQueue[res];
		bf[0][j] += cp;			//O->t.
		i = bPrev[ j+pSize-1];	//bug. i->j
		bf[ i+1][j] += cp;		//P->O.
		j = bPrev[i];

		/*s*/
		thr_stat_v.path_length_sum += 2;
		/*s*/

		while( j != 0)			
		{	
			bf[ i+1][ j-pSize+1] -= cp;	//P->O.
			i = bPrev[j];
			bf[ i+1][ j-pSize+1] += cp;	//P->O.
			j = bPrev[i];

			thr_stat_v.path_length_sum += 2;
		}
		bf[ i+1][0] += cp;				//

		res = find_path_gen( );

		/*s*/
		thr_stat_v.n_path ++;
		/*s*/
	}//while(res)

	//statistic info.


	//Collect the flow information.
	flow = 0;
	for( i=1; i<=pSize; i++)
	{
		flow += bf[i][0];
	}

	return flow;
}


/*
* test_ford_fulkerson tests the correctness of the ford_fulkerson_gen algorithm.
*/
void test_ford_fulkerson( )
{
	int i, j, mf, n_row, n_col;
	FILE *c_fp, *i_fp, *o_fp;
	c_fp = fopen( "config.txt", "r");
	i_fp = fopen( "data.txt", "r");
	o_fp = fopen( "output.txt", "w");
	if( c_fp == NULL || i_fp == NULL || o_fp == NULL)
	{
		fprintf( stderr, "The config/input/output file cannot be open.\n");
		exit(0);
	}
	
	//Reading the data.
	fscanf( c_fp, "%i%i", &pSize, &oSize);			
	n_row = pSize + 1;						//pSize: the # of subset 1(excluding the source).
	n_col = oSize + 1;						//oSize: the # of subset 2(excluding the sink).
	for( i=0; i<n_row; i++)						
	{
		for( j=0; j<n_col; j++)
		{
			fscanf( i_fp, "%i", &bGraph[i][j]);
		}
	}

	//Performing the maximal flow procedure.
	ford_fulkerson_gen( );

	//Printing the results.
	mf = 0;
	for( i=1; i<n_row; i++)
		mf += bf[i][0];
	fprintf( o_fp, "Maximal flow: %i\n", mf);

	for( i=0; i<n_row; i++)
	{
		for( j=0; j<n_col; j++)
		{
			fprintf( o_fp, "%i  ", bf[i][j]);
		}
		fprintf( o_fp, "\n");
	}

	//Releasing the reources.
	fclose( c_fp);
	fclose( i_fp);
	fclose( o_fp);
}


/*
* ford_fulkerson_bi is used to find the Maximum Flow with 'find_path_gen'( for Algorithm 2)
* 
* @Param/Return void.
*/
void ford_fulkerson_bi( )
{
	int i, j, res;
	//bf array is initialized in 'IniGraph3' function--Necessary for 'Preprocess' functions.

	res = find_path_gen( );
	while( res != 0)	//(int**)graph, (int**)f, (int*&)prev,
	{
		//Specify the cp value(needn't when unweighted O set, that is cp = 1).
		//Update the augmenting path.
		j = bQueue[res];
		bf[0][j] = 1;			//O->t.
		i = bPrev[j+pSize-1];	//bug. i->j
		bf[i+1][j] = 1;			//P->O.
		j = bPrev[i];
		while( j != 0)			
		{	
			bf[ i+1][j-pSize+1] = 0;		//P->O.
			i = bPrev[j];
			bf[ i+1][j-pSize+1] = 1;		//P->O.
			j = bPrev[i];
		}
		// j==0
		bf[ i+1][0] += 1;
		res = find_path_gen( );
	}//while(res)
}

/*------------------------------Supporting APIs------------------------------------*/

/*
* partition & quick_sort is used to the distArray.
* 
* @Param d, p1, p2 indicates the array, start and end respectively.
*
* @Return mid(int) for 'partition' and void for 'quick_sort'.
*
int partition( DOUBLE* d, int p1, int p2)
{
	if( p1 == p2)				//Couldn't be 'p1 > p2'.
		return p1;
	DOUBLE tmp = d[p1];
	while( p1 < p2)
	{
		while( p1 < p2 && d[p2] >= tmp)
			p2 --;
		if( p1 != p2)
			d[p1++] = d[p2];
		while( p1 < p2 && d[p1] <= tmp)
			p1 ++;
		if( p1 != p2)
			d[p2--] = d[p1];
	}
	d[p1] = tmp;
	return p1;
}

void quick_sort( DOUBLE* d, int p1, int p2)	//Need test.
{
	if( p1 >= p2)
		return;
	int mid = partition( d, p1, p2);

	quick_sort( d, p1, mid-1);
	quick_sort( d, mid+1, p2);
}*/

/*---------------------Algorithm 2 (Advance--General edition)-----------------------*/

/*
* Same to other IniGraph functions.
*/
void flow_network_ini( DOUBLE cri)
{
	int i,j;

	//Initialize the link array.
	for( i=0; i<pSize; i++)
		pWeight[i] = 0;
	for( i=0; i<oSize; i++)
		oWeight[i] = 0;
	
	//Initialize the bGraph structure.
	bGraph[0][0] = 0;
	for( i=1; i<=pSize; i++)			//s->P.
	{
		bGraph[i][0] = pCapacity[i-1];

		for( j=1; j<=oSize; j++)		//P->O.
		{
			if( dist[i-1][j-1] <= cri)	//Reserve.
			{
				bGraph[i][j] = pCapacity[i-1] <= oCapacity[j-1] ? pCapacity[i-1] : oCapacity[j-1];
				pWeight[i-1] += bGraph[i][j];
				oWeight[j-1] += bGraph[i][j];
			}
			else						//Omit.
			{
				bGraph[i][j] = 0;
			}
		}//for(j)
	}//for(i)

	for( j=1; j<=oSize; j++)			//O->t.
	{
		bGraph[0][j] = oCapacity[j-1];
	}
	
	//Initialize bf array.
	for( i=0; i<=pSize; i++)
	{
		for( j=0; j<=oSize; j++)
		{
			bf[i][j] = 0;
		}//for(j)
	}//for(i)

	return;
}

/*
* flow_network_ini_inc initializes the network flow in an incremental way.
*
* @cri indicates the criteria.
* @change_tag indicates the change direction of the criteria.
* 1 for larger; -1 for smaller.
*/
void flow_network_ini_inc( DOUBLE cri, int change_tag)
{
	int i, j;
/*
	memset( pWeight, 0, sizeof(int) * pSize);
	memset( oWeight, 0, sizeof(int) * oSize);
*/
	//Initializing the graph.
	for( i=1; i<=pSize; i++)
	{
		for( j=1; j<=oSize; j++)
		{
			if( change_tag == 1)	//cri is larger.
			{
				if( bGraph[i][j] == 0 && dist[i-1][j-1] <= cri)	//Only this case needed to be handled.
				{
					bGraph[i][j] = pCapacity[i-1] <= oCapacity[j-1] ? 
								pCapacity[i-1] : oCapacity[j-1];
					//pWeight[i-1] += bGraph[i][j];
					oWeight[j-1] += bGraph[i][j];
				}
			}
			else	//change_tag == -1: cri is smaller.
			{
				if( bGraph[i][j] !=0 && dist[i-1][j-1] > cri)
				{
					if( bf[i][j] != 0)
					{
						bf[i][0] -= bf[i][j];
						bf[0][j] -= bf[i][j];
						bf[i][j] = 0;
					}
					oWeight[j-1] -= bGraph[i][j];
					bGraph[i][j] = 0;
				}
			}
		}//for(j)
	}//for(i)
}

/*
 * max_flow_pruning is used to precess on the o_weight array.
 *
 * @Param/Return void.
 */
bool max_flow_pruning( )
{
	int i;
	for( i=0; i<oSize; i++)
	{
		if( oWeight[i] < oCapacity[i])			//This customer couldn't be satisfied.
			return false;
/*		else if( oWeight[i] == oCapacity[i])	//Make the 'Must' assignment.
		{
			bf[0][i+1] = oWeight[i];					//o->t.
			for( j=1; j<=pSize; j++)
			{
				if( bGraph[j][i+1] != 0)
				{
					//bf[0][i+1] += bGraph[j][i+1];		//o->t.
					bf[j][i+1] = bGraph[j][i+1];		//p->o.
					pWeight[j-1] -= bGraph[j][i+1];
					bGraph[j][i+1] = 0;
					bf[j][0] += bf[j][i+1];				//s->p.
					if( bGraph[j][0] - bf[j][0] < 0)	//bug.
						return false;				
					if( bf[0][i+1] == oCapacity[i])
						break;
				}
			}//for(j).
		}
*/
	}//for(i).
	return true;
}

/*
* PreProcess_Sub5 is used to precess on the p_weight array.
*
* @Param/Return void.
*
void PreProcess_Sub5( )
{
	int i, j, cp;
	for( i=0; i<pSize; i++)
	{
		if( pCapacity[i] - bf[i+1][0] >= pWeight[i] && pWeight[i] != 0)
		{
			//bf[i+1][0] += pWeight[i];				//s->p--bug.
			bGraph[i+1][0] = 0;
			for( j=1; j<=oSize; j++)
			{
				if( bGraph[i+1][j] != 0 && bGraph[0][j] - bf[0][j] > 0)
				{
					cp = bGraph[i+1][j] < bGraph[0][j] - bf[0][j] ? 
						bGraph[i+1][j] : bGraph[0][j] - bf[0][j];
					bf[i+1][0] += cp;
					bf[i+1][j] += cp;				//p->o.
					bGraph[i+1][j] = 0;				//
					bf[0][j] += cp;					//o->t.
				}
			}//for(j).
		}
	}//for(i).
}*/

/*
* max_flow_pre_matching is used to precess in a random way.
*
* @Param/Return void.
*/
void max_flow_pre_matching( )
{
	int i, j, cp = INT_MAX;
	for( i=1; i<=pSize; i++)
	{
		if( bGraph[i][0] - bf[i][0] <= 0)
			continue;
		for( j=1; j<=oSize; j++)
		{
			if( bGraph[i][j] - bf[i][j] > 0 && bGraph[0][j] - bf[0][j] > 0)
			{
				//Specify the cp value.
				cp = bGraph[i][0] - bf[i][0] < bGraph[i][j] - bf[i][j] ? bGraph[i][0] - bf[i][0] : bGraph[i][j] - bf[i][j];
				if( cp > bGraph[0][j] - bf[0][j])
					cp = bGraph[0][j] - bf[0][j];

				//Construct an path temporarily.
				bf[i][0] += cp;
				bf[i][j] += cp;
				bf[0][j] += cp;
				
				if( bGraph[i][0] - bf[i][0] <= 0)
					break;
			}//
		}//for(j).
	}//for(i).
}

/*
 *	The implementation of the Incremental Breadth First Search algorithm (IBFS).
 */
int IBFS( )
{
	int i, j, flow;
	//int node_n, edge_n;
	typedef IBFSGraph<int,int,int> IBFSGraph;

	IBFSGraph* g;

	//Prepare the flow network based on bGraph.
	g = new IBFSGraph( pSize + oSize, pSize * oSize); //(estimated # of nodes, estimated # of edges);

	//g -> add_node( 1); 
	//g -> add_node( 1); 

	for( i=0; i<pSize+oSize; i++)
		g->add_node( i+1);

	//The P objects are identified as 0, 1, ..., pSize - 1.
	for( i=0; i<pSize; i++)
	{
		g->add_tweights( i, pCapacity[ i], 0);
	}

	//The O objects are identified as pSize + 0, pSize + 1, ..., pSize + oSize - 1.
	for( j=0; j<oSize; j++)
	{
		g->add_tweights( pSize + j, 0, oCapacity[ j]);
	}

	//
	for( i=1; i<=pSize; i++)
	{
		for( j=1; j<=oSize; j++)
		{
			if( bGraph[ i][ j] > 0)
			{
				//Add edge (i-1, pSize + j-1);
				g->add_edge( i - 1, pSize + j - 1, pSize + j - 1, bGraph[ i][ j]);
			}
		}
	}
	
	
	//g -> add_tweights( 0, 1, 5 ); //(sourceCapacity, sinkCapacity)
	//g -> add_tweights( 1, 2, 6 ); //(sourceCapacity, sinkCapacity)
	//g -> add_edge( 0, 1, 3, 4 ); (capacity, reverse capacity)


	//Compute the maximum-flow.
	flow = g -> maxflow();

	//printf("Flow = %d\n", flow);

	delete g;

	return flow;
}

//The implementation of the Boykov-Kolmogorov (BK) algorithm.
int BK( )
{
	int i, j, flow;
	//int node_n, edge_n;
	typedef Graph<int,int,int> GraphType;

	GraphType* g;

	//Prepare the flow network based on bGraph.
	g = new GraphType( pSize + oSize, pSize * oSize); //(estimated # of nodes, estimated # of edges);

	//g -> add_node( 1); 
	//g -> add_node( 1); 
	for( i=0; i<pSize+oSize; i++)
		g->add_node( i + 1);

	//The P objects are identified as 0, 1, ..., pSize - 1.
	for( i=0; i<pSize; i++)
	{
		g->add_tweights( i, pCapacity[ i], 0);
	}

	//The O objects are identified as pSize + 0, pSize + 1, ..., pSize + oSize - 1.
	for( j=0; j<oSize; j++)
	{
		g->add_tweights( pSize + j, 0, oCapacity[ j]);
	}

	//
	for( i=1; i<=pSize; i++)
	{
		for( j=1; j<=oSize; j++)
		{
			if( bGraph[ i][ j] > 0)
			{
				//Add edge (i-1, pSize + j-1);
				g->add_edge( i - 1, pSize + j - 1, pSize + j - 1, bGraph[ i][ j]);
			}
		}
	}
	
	
	//g -> add_tweights( 0, 1, 5 ); //(sourceCapacity, sinkCapacity)
	//g -> add_tweights( 1, 2, 6 ); //(sourceCapacity, sinkCapacity)
	//g -> add_edge( 0, 1, 3, 4 ); (capacity, reverse capacity)


	//Compute the maximum-flow.
	flow = g -> maxflow();

	//printf("Flow = %d\n", flow);

	delete g;

	return flow;
}

/*
* Inc_IsMS_Advance is the incremental version of the maximum flow.
*
* @pTag indicates the Preprocesing steps.
* @cri indicates the distance criteria for the construction.
* speed_opt = 0: No preprocessing step.
* speed_opt = 1; Increamental manner.
* speed_opt = 2; Incremental manner + Prunning.
* speed_opt = 3; Incremental manner + Random.
* speed_opt = 4; Incremental manner + Prunning + Random.
* @change_tag indicates the change direction of the criteria of the distance.
*
* @Return true if maximum flow can satisfy all the customers; otherwise, return false.
*/
bool comp_max_flow( int alg_opt, int speed_opt, DOUBLE cri, int change_tag)
{
	int flow;
/**/
	if( speed_opt-- > 0 && change_tag != 0)	//Incremental manner.
		flow_network_ini_inc( cri, change_tag);
	else
		flow_network_ini( cri);
	if( speed_opt-- > 0)						//Prunning.
	{	
		if( !max_flow_pruning( ))
			return false;
	}
	if( speed_opt-- > 0)						//Random.
	{
		max_flow_pre_matching( );
	}
/**/
/*
	if( speed_opt == 0 || change_tag == 0)
	{
		flow_network_ini( cri);
	}
	else								//pTag > 0 && chang_tag != 0
	{
		//pTag != 0 && change_tag != 0.
		flow_network_ini_inc( cri, change_tag);

		if( speed_opt == 2)
		{
			if( !max_flow_pruning( ))
				return false;
		}
		else if( speed_opt == 3)
		{
			max_flow_pre_matching( );
		}
		else if( speed_opt == 4)
		{
			if( !max_flow_pruning( ))
				return false;

			max_flow_pre_matching( );
		}
	}
*/
	//Performing the maximum flow--Ford-Fulkerson algorithm.
	if( alg_opt == 1)
		flow = ford_fulkerson_gen( );
	else if( alg_opt == 2)
		flow = IBFS( );	
	else if( alg_opt == 3)
		flow = BK( );
	
	if( flow == o_w_sum)
		return true;
	else 
		return false;
}

/*
* Threshold_Adapt is modified to support the incremental manner from "SPM_MM_Advance".
*
* @alg_opt
*
* @speed_opt
*	
*/	
DOUBLE Threshold_Adapt( int alg_opt, int speed_opt)
{
	int ini_tag;
	long left, right, mid;

#ifndef WIN32
	get_cur_time( &(preprocess_start));
#endif

	ini_tag = 0;
	left = 0;
	right = pSize * oSize - 1;

	printf( "Threshold-Adapt ...\n");
	if( alg_opt == 1)
	{
		printf( "max-flow: Augmenting Path (speed_opt = %i)\n", speed_opt);
	}
	else if( alg_opt == 2)
	{
		printf( "max-flow: IBFS (speed_opt = %i)\n", speed_opt);
	}
	else if( alg_opt == 3)
	{
		printf( "max-flow: BK (speed_opt = %i)\n", speed_opt);
	}

	//Calculate the distance array.
	comp_dist_matrix( );
		
	//Sort the distance array.
	//quick_sort( distArray, 0, oSize*pSize-1);
	qsort( distArray, pSize * oSize, sizeof( DOUBLE), DOUBLE_comparison);

#ifndef WIN32
	get_cur_time( &(preprocess_end));
#endif
		
	while( left < right)		//Bisection Method.
	{
		mid = (left+right)/2;

		//t
		printf( "Checking %f:\t", distArray[ mid]);
		//t

		if( comp_max_flow( alg_opt, speed_opt, distArray[mid], ini_tag))
		{
			//t
			printf( "1\n");
			//t

			right = mid;
			ini_tag = -1;
		}
		else
		{
			//t
			printf( "0\n");
			//t

			left = mid + 1;
			ini_tag = 1;
		}
	}//while

#ifndef WIN32
	get_cur_time( &(match_end));
#endif

	return distArray[right];
}



/*---------------------------Added functions--------------------------------*/

/*
* TA_read_config reads the configuration--emp_config_t.
*/
void TA_read_config( )
{
	FILE* c_fp = fopen( EMP_CONFIG_FILE, "r");
	if( c_fp == NULL)
	{
		fprintf( stderr, "The config.txt file cannot be open.\n");
		exit(0);
	}

	//Reading the information.
	fscanf( c_fp, "%s%s", _cfg.o_file, _cfg.p_file);
	fscanf( c_fp, "%i%i", &(_cfg.n_o), &( _cfg.n_p));
	fscanf( c_fp, "%i", &(_cfg.dim));
	//fscanf( c_fp, "%f%f", &_cfg.n_r, &_cfg.w_r);

	//fscanf( c_fp, "%i%i%i%i", &_cfg.thr_opt, &_cfg.speed_opt, &_cfg.swap_opt, &_cfg.graph_tag);
	fscanf( c_fp, "%i%i", &_cfg.thr_opt, &_cfg.swap_opt);

	_cfg.speed_opt = 2;

	//Releasing the related the resource.
	fclose( c_fp);
}

/*
* TA_read_data works corresponding to Read_Data.
*/
void TA_read_data( )
{
	int i, j;
	char filter[10];
	p_w_sum = 0, o_w_sum = 0;

	FILE* p_fp, *o_fp;

	p_fp = fopen( _cfg.p_file, "r");
	o_fp = fopen( _cfg.o_file, "r");
	if( p_fp == NULL || o_fp == NULL)
	{
		fprintf( stderr, "The P/O objects file cannot be open.\n");
		exit( 0);
	}

	//Reading the P objects related data.
	pSize = _cfg.n_p;
	
	pPoint = (_point_t*)malloc( pSize * sizeof(_point_t));
	pCapacity = ( int*)malloc( pSize * sizeof( int));

	for( i=0; i<pSize; i++)
	{
		fscanf( p_fp, "%s", filter);						//filter the "i:".		
		fscanf( p_fp, "%i", pCapacity + i);


	//	if( DISCRITITIZE_TAG)
	//		pCapacity[ i] -= pCapacity[ i] % DISCRITITIZE_SCALE;

		if( pCapacity[ i] == 0)
			pCapacity[ i] = 1;
/*t*
		pCapacity[i] = _cfg.n_r * _cfg.w_r;
/*t*/
		pPoint[i].ord = ( float*)malloc( sizeof( float) * _cfg.dim);

		for( j=0; j<_cfg.dim; j++)
		{
			fscanf( p_fp, "%f", &(pPoint[i].ord[j]));

			if( PERTUBATION_TAG == 1)
				pPoint[ i].ord[ j] += ( float)rand_f( PERTUBATION_INTERVAL_LEFT, PERTUBATION_INTERVAL_RIGHT);

		}
		p_w_sum += pCapacity[i];
	}

	//Reading the O objects related data.
	oSize = _cfg.n_o;

	oPoint = ( _point_t*)malloc( oSize * sizeof( _point_t));
	oCapacity = ( int*)malloc( oSize * sizeof( int));

	for( i=0; i<oSize; i++)
	{
		fscanf( o_fp, "%s", filter);							//filter the "i:".
		fscanf( o_fp, "%i", oCapacity+i);

		if( DISCRITITIZE_TAG)
			oCapacity[ i] -= oCapacity[ i] % DISCRITITIZE_SCALE;

		if( oCapacity[ i] == 0)
			oCapacity[ i] = 1;
/*t*
		oCapacity[i] = 1;
/*t*/
		oPoint[i].ord = ( float*)malloc( sizeof(float) * _cfg.dim);

		for( j=0; j<_cfg.dim; j++)
		{
			fscanf( o_fp, "%f", &(oPoint[i].ord[j]));

			if( PERTUBATION_TAG == 1)
				oPoint[ i].ord[ j] += ( float)rand_f( PERTUBATION_INTERVAL_LEFT, PERTUBATION_INTERVAL_RIGHT);
			
		}
		
		o_w_sum += oCapacity[i];
	}

	//
	//Testing the assumption condition.
	if( p_w_sum < o_w_sum)
	{
		fprintf( stderr, "The assumption that p_w_sum is at leat o_w_sum is not satisfied.\n");
		exit(0);
	}


	_cfg.n_r = ( float)_cfg.n_o / _cfg.n_p;
	_cfg.w_r = ( float)p_w_sum / o_w_sum;


}
/*
* TA_calc_dist calculates the distance between two points.
*
* @Param ord1, ord2 indicate the two coordinates respectively.
* 
* @Return the distance's square value.
*/
DOUBLE TA_calc_dist( const float* ord1, const float* ord2)
{
	int i;
	DOUBLE dist = 0;

	for( i=0; i<_cfg.dim; i++)
	{
		dist += (ord1[i] - ord2[i]) * (ord1[i] - ord2[i]);
	}

	return (DOUBLE)sqrt( dist);
}

/*
* comp_dist_matrix is used to calculate the distance array of all possible matching.
*
* @Param/Return void.
*/
void comp_dist_matrix( )
{
	int i, j, tmp = 0;
	DOUBLE res;

	for( i=0; i<pSize; i++)			//Travel set P.
	{
		for( j=0; j<oSize; j++)		//Travel set O.
		{
			res = TA_calc_dist( pPoint[i].ord, oPoint[j].ord);

			distArray[tmp++] = res;
			dist[i][j] = res;
		}		
	}

	return;
}

/*
* TA_data_alloc allcates the main memory for the graph related data.
*/
void TA_data_alloc( )
{
	int i;

	bGraph = ( int**)malloc( sizeof(int*) * (pSize + 1));
	memset( bGraph, 0, sizeof( int*) * ( pSize + 1));

	bf = ( int**)malloc( sizeof( int*) * (pSize + 1));
	memset( bf, 0, sizeof( int*) * ( pSize + 1));

	/*s*/
	thr_stat_v.memory_v += sizeof( float*) * (pSize + 1) * 2;
	/*s*/
		
	for( i=0; i<pSize + 1; i++)
	{
		bGraph[i] = (int*)malloc( sizeof(int) * (oSize + 1));
		memset( bGraph[ i], 0, sizeof( int) * ( oSize + 1));

		bf[i] = (int*)malloc( sizeof(int) * (oSize + 1));
		memset( bf[ i], 0, sizeof( int) * ( oSize + 1));
	}

	/*s*/
	thr_stat_v.memory_v += ( pSize + 1) * 2 * sizeof( float) * (oSize + 1);
	/*s*/

	bPrev = (int*)malloc( sizeof(int) * (pSize + oSize + 2));
	bQueue = (int*)malloc( sizeof(int) * (pSize + oSize + 2));
	bTag = (int*)malloc( sizeof(int) * (pSize + oSize + 2));

	pLink = (int*)malloc( sizeof(int) * pSize);
	oLink = (int*)malloc( sizeof(int) * oSize);

	pWeight = (int*)malloc( sizeof(int) * pSize);
	oWeight = (int*)malloc( sizeof(int) * oSize);

	/*s*/
	thr_stat_v.memory_v += sizeof(int) * ( pSize + oSize + 2) * 3 + 
		sizeof( int) * ( pSize + oSize) * 2;
	/*s*/

	dist = (DOUBLE**)malloc( sizeof(DOUBLE*) * pSize);
	for( i=0; i<pSize; i++)
		dist[i] = (DOUBLE*)malloc( sizeof(DOUBLE) * oSize);
	distArray = ( DOUBLE*)malloc( sizeof(DOUBLE) * (pSize * oSize));

	/*s*/
	thr_stat_v.memory_v += sizeof( DOUBLE) * pSize * oSize;		//dist & distArray.
	
	if( thr_stat_v.memory_v > thr_stat_v.memory_max)
		thr_stat_v.memory_max = thr_stat_v.memory_v;
	/*s*/
}

/*
* TA_data_release releases the graph related data.
*/
void TA_data_release( )
{
	int i;
	for( i=0; i<pSize+1; i++)
	{
		free( bGraph[i]);
		free( bf[i]);
	}

	free( bGraph);
	free( bf);

	/*s*/
	thr_stat_v.memory_v -= sizeof( float*) * (pSize + 1) * 2;
	thr_stat_v.memory_v -= (pSize + 1) * 2 * sizeof( float) * (oSize + 1);
	/*s*/

	free( bPrev);
	free( bQueue);
	free( bTag);
	free( pLink);
	free( oLink);
	free( pWeight);
	free( oWeight);
	
	/*s*/
	thr_stat_v.memory_v -= sizeof(int) * (pSize + oSize + 2) * 3 + 
		sizeof( int) * ( pSize + oSize) * 2;
	/*s*/


	for( i=0; i<pSize; i++)
		free( dist[i]);
	free( dist);
	free( distArray);

	/*s*/
	thr_stat_v.memory_v -= sizeof( DOUBLE) * pSize * oSize;		//dist & distArray.	
	/*s*/
}

/*
 * emp_Threshold_Adapt tests the implementation of "Inc_SPM_Advance".
 * 
 * @pTag indicates the preprocessing steps.
 */
void emp_Threshold_Adapt( )
{
	DOUBLE mmd;

	/*s*/
	memset( &thr_stat_v, 0, sizeof(stat_t));
	/*s*/

	//Reading the config.
	//printf( "Reading configuration ...\n");
	TA_read_config( );

	//Reading the raw data.
	printf( "Reading data...\n");
	TA_read_data( );

	//s
	thr_stat_v.memory_v += pSize * sizeof( _point_t) + pSize * sizeof( int) + pSize * sizeof( float) * _cfg.dim +
							oSize * sizeof( _point_t) + oSize * sizeof( int) + oSize * sizeof( float) * _cfg.dim;
	if( thr_stat_v.memory_max > thr_stat_v.memory_v)
		thr_stat_v.memory_max = thr_stat_v.memory_v;
	//s

	//Allocating the main memory related to the grapph dynamically.
	//printf( "Allocating data...\n");
	TA_data_alloc( );

	//Performing the SPM-MM calculation.
	//printf( "Threshold-Adapt (opt = %i) ...\n", _cfg.thr_opt);

	mmd = Threshold_Adapt( _cfg.thr_opt, _cfg.speed_opt);
	
#ifndef	WIN32 	
	float tmp;
	collect_time( &(preprocess_start), &(preprocess_end), &thr_stat_v.preprocess_time, &tmp);
	collect_time( &(preprocess_end), &(match_end), &thr_stat_v.matching_time, &tmp);
#endif

	//time.
	thr_stat_v.total_time = thr_stat_v.preprocess_time + thr_stat_v.matching_time;

	//augmenting path & queue & expension.
	if( thr_stat_v.n_path != 0)
	{
		thr_stat_v.path_length_aver = thr_stat_v.path_length_sum / thr_stat_v.n_path;
		thr_stat_v.que_length_aver = thr_stat_v.que_length_sum / thr_stat_v.n_path;
	}


	//s
	thr_stat_v.memory_v -= pSize * sizeof( _point_t) + pSize * sizeof( int) + pSize * sizeof( float) * _cfg.dim +
							oSize * sizeof( _point_t) + oSize * sizeof( int) + oSize * sizeof( float) * _cfg.dim;
	//s


	//Print the statistics.
	print_stat( mmd, thr_stat_v);
		
	//Releasing the related resource.
	TA_data_release( );		//graph
	
	free( oPoint);
	free( pPoint);
	free( oCapacity);
	free( pCapacity);

	printf( "mmd: %lf\n", mmd);
	printf( "total time: %f\n", thr_stat_v.total_time);
	//printf( "Memory balance: %f\n", thr_stat_v.memory_v);
	printf( "Memory (MB): %f\n", thr_stat_v.memory_max / ( 1024 * 1024));
}